"""Testing SMILES dataset with lazy backend."""
import json
import os
import unittest

import numpy as np
from torch.utils.data import DataLoader

from pytoda.datasets import SMILESTokenizerDataset
from pytoda.smiles import SMILESTokenizer, metadata
from pytoda.smiles.smiles_language import TOKENIZER_CONFIG_FILE
from pytoda.tests.utils import TestFileContent

CONTENT = os.linesep.join(
    ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
)
MORE_CONTENT = os.linesep.join(
    [
        'COCC(C)N	CHEMBL3184692',
        'COCCOC	CHEMBL1232411',
        'O=CC1CCC1	CHEMBL18475',  # longest with length 9
        'NC(=O)O	CHEMBL125278',
    ]
)
LONGEST = 9


class TestSMILESTokenizerDatasetEager(unittest.TestCase):
    """Testing SMILES dataset with eager backend."""

    def setUp(self):
        self.backend = 'eager'
        print(f'backend is {self.backend}')
        self.content = CONTENT
        self.other_content = MORE_CONTENT
        self.longest = LONGEST

    def test___len__(self) -> None:
        """Test __len__."""

        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:
                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                )
                self.assertEqual(len(smiles_dataset), 8)

    def test___getitem__(self) -> None:
        """Test __getitem__."""

        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:
                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=True,
                    augment=False,
                    kekulize=True,
                    sanitize=True,
                    all_bonds_explicit=True,
                    remove_chirality=True,
                    backend=self.backend,
                )
                pad_index = smiles_dataset.smiles_language.padding_index
                start_index = smiles_dataset.smiles_language.start_index
                stop_index = smiles_dataset.smiles_language.stop_index
                c_index = smiles_dataset.smiles_language.token_to_index['C']
                o_index = smiles_dataset.smiles_language.token_to_index['O']
                n_index = smiles_dataset.smiles_language.token_to_index['N']
                s_index = smiles_dataset.smiles_language.token_to_index['S']
                d_index = smiles_dataset.smiles_language.token_to_index['-']

                sample = 0
                padding_len = smiles_dataset.smiles_language.padding_length - (
                    len(
                        # str from underlying concatenated _smi dataset
                        smiles_dataset.dataset[sample]
                    )
                    * 2
                    - 1  # just d_index, no '=', '(', etc.
                )
                self.assertListEqual(
                    smiles_dataset[sample].numpy().flatten().tolist(),
                    [pad_index] * padding_len
                    + [c_index, d_index, c_index, d_index, o_index],
                )

                sample = 3
                padding_len = smiles_dataset.smiles_language.padding_length - (
                    len(
                        # str from underlying concatenated _smi dataset
                        smiles_dataset.dataset[sample]
                    )
                    * 2
                    - 1  # just d_index, no '=', '(', etc.
                )
                self.assertListEqual(
                    smiles_dataset[sample].numpy().flatten().tolist(),
                    [pad_index] * padding_len
                    + [n_index, d_index, c_index, d_index, c_index, d_index, s_index],
                )

                sample = 5
                padding_len = smiles_dataset.smiles_language.padding_length - (
                    len(
                        # str from underlying concatenated _smi dataset
                        smiles_dataset.dataset[sample]
                    )
                    * 2
                    - 1  # just d_index, no '=', '(', etc.
                )
                self.assertListEqual(
                    smiles_dataset[sample].numpy().flatten().tolist(),
                    [pad_index] * padding_len
                    + [
                        c_index,
                        d_index,
                        o_index,
                        d_index,
                        c_index,
                        d_index,
                        c_index,
                        d_index,
                        o_index,
                        d_index,
                        c_index,
                    ],
                )

                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    backend=self.backend,
                )
                c_index = smiles_dataset.smiles_language.token_to_index['C']
                o_index = smiles_dataset.smiles_language.token_to_index['O']

                self.assertListEqual(
                    smiles_dataset[0].numpy().flatten().tolist(),
                    [c_index, c_index, o_index],
                )
                self.assertListEqual(
                    smiles_dataset[5].numpy().flatten().tolist(),
                    [c_index, o_index, c_index, c_index, o_index, c_index],
                )

                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    add_start_and_stop=True,
                    backend=self.backend,
                )
                pad_index = smiles_dataset.smiles_language.padding_index
                start_index = smiles_dataset.smiles_language.start_index
                stop_index = smiles_dataset.smiles_language.stop_index
                c_index = smiles_dataset.smiles_language.token_to_index['C']
                o_index = smiles_dataset.smiles_language.token_to_index['O']

                self.assertEqual(
                    smiles_dataset.smiles_language.padding_length, self.longest + 2
                )

                sample = 0
                self.assertEqual(smiles_dataset.dataset[sample], 'CCO')
                padding_len = (
                    smiles_dataset.smiles_language.padding_length
                    - (
                        len(
                            # str from underlying concatenated _smi dataset
                            smiles_dataset.dataset[sample]
                        )
                    )
                    - 2
                )  # start and stop
                self.assertListEqual(
                    smiles_dataset[sample].numpy().flatten().tolist(),
                    [pad_index] * padding_len
                    + [start_index, c_index, c_index, o_index, stop_index],
                )

                sample = 5
                self.assertEqual(smiles_dataset.dataset[sample], 'COCCOC')
                padding_len = (
                    smiles_dataset.smiles_language.padding_length
                    - (
                        len(
                            # str from underlying concatenated _smi dataset
                            smiles_dataset.dataset[sample]
                        )
                    )
                    - 2
                )  # start and stop
                self.assertListEqual(
                    smiles_dataset[sample].numpy().flatten().tolist(),
                    [pad_index] * padding_len
                    + [
                        start_index,
                        c_index,
                        o_index,
                        c_index,
                        c_index,
                        o_index,
                        c_index,
                        stop_index,
                    ],
                )

                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    augment=True,
                    backend=self.backend,
                )
                np.random.seed(0)
                for randomized_smiles in ['C(S)CN', 'NCCS', 'SCCN', 'C(N)CS', 'C(CS)N']:
                    token_indexes = smiles_dataset[3].numpy().flatten().tolist()
                    smiles = smiles_dataset.smiles_language.token_indexes_to_smiles(
                        token_indexes
                    )
                    self.assertEqual(smiles, randomized_smiles)

                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=True,
                    remove_bonddir=True,
                    selfies=True,
                    backend=self.backend,
                )
                c_index = smiles_dataset.smiles_language.token_to_index['[C]']
                o_index = smiles_dataset.smiles_language.token_to_index['[O]']
                n_index = smiles_dataset.smiles_language.token_to_index['[N]']
                s_index = smiles_dataset.smiles_language.token_to_index['[S]']

                self.assertListEqual(
                    smiles_dataset[0].numpy().flatten().tolist(),
                    [start_index, c_index, c_index, o_index, stop_index],
                )
                self.assertListEqual(
                    smiles_dataset[3].numpy().flatten().tolist(),
                    [start_index, n_index, c_index, c_index, s_index, stop_index],
                )

    def test_data_loader(self) -> None:
        """Test data_loader."""
        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:
                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                )
                data_loader = DataLoader(smiles_dataset, batch_size=4, shuffle=True)
                for batch_index, batch in enumerate(data_loader):
                    self.assertEqual(batch.shape, (4, self.longest))
                    if batch_index > 10:
                        break

    def _test_indexed(self, ds, keys, index):
        key = keys[index]
        positive_index = index % len(ds)
        # get_key (support for negative index?)
        self.assertEqual(key, ds.get_key(positive_index))
        self.assertEqual(key, ds.get_key(index))
        # get_index
        self.assertEqual(positive_index, ds.get_index(key))
        # get_item_from_key
        self.assertTrue(all(ds[index] == ds.get_item_from_key(key)))
        # keys
        self.assertSequenceEqual(keys, list(ds.keys()))
        # duplicate keys
        self.assertFalse(ds.has_duplicate_keys)

    def test_all_base_for_indexed_methods(self):

        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:
                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                )
                smiles_dataset_0 = SMILESTokenizerDataset(
                    a_test_file.filename, backend=self.backend
                )
                smiles_dataset_1 = SMILESTokenizerDataset(
                    another_test_file.filename, backend=self.backend
                )
        all_smiles, all_keys = zip(
            *(
                pair.split('\t')
                for pair in self.content.split(os.linesep)
                + self.other_content.split(os.linesep)
            )
        )

        for ds, keys in [
            (smiles_dataset, all_keys),
            (smiles_dataset_0, all_keys[:4]),
            (smiles_dataset_1, all_keys[4:]),
            # no transformation on
            # concat delegation to _SmiLazyDataset/_SmiEagerDataset
            (smiles_dataset_0 + smiles_dataset_1, all_keys),
        ]:
            index = -1
            self._test_indexed(ds, keys, index)

        # duplicate
        duplicate_ds = smiles_dataset_0 + smiles_dataset_0
        self.assertTrue(duplicate_ds.has_duplicate_keys)

        # SMILESTokenizerDataset tests and raises
        with TestFileContent(self.content) as a_test_file:
            with self.assertRaises(KeyError):
                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename, a_test_file.filename, backend=self.backend
                )

    def test_pretrained__getitem__(self) -> None:
        """Test __getitem__."""
        pretrained_path = os.path.join(
            os.path.dirname(os.path.abspath(metadata.__file__)),
            'tokenizer',
        )
        smiles_language = SMILESTokenizer.from_pretrained(pretrained_path, padding=True)

        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:

                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    smiles_language=smiles_language,
                    iterate_dataset=False,
                )

                config_file = os.path.join(pretrained_path, TOKENIZER_CONFIG_FILE)
                with open(config_file, encoding="utf-8") as fp:
                    max_length = json.load(fp)['max_token_sequence_length']
            pad_index = smiles_dataset.smiles_language.padding_index
            c_index = smiles_dataset.smiles_language.token_to_index['C']
            o_index = smiles_dataset.smiles_language.token_to_index['O']

            self.assertEqual(max_length, smiles_dataset.smiles_language.padding_length)
            sample = 0
            padding_len = smiles_dataset.smiles_language.padding_length - (
                len(
                    # str from underlying concatenated _smi dataset
                    smiles_dataset.dataset[sample]
                )
            )
            self.assertListEqual(
                smiles_dataset[sample].numpy().flatten().tolist(),
                [pad_index] * padding_len + [c_index, c_index, o_index],
            )

        # Test case where the language is adapted
        smiles_language = SMILESTokenizer.from_pretrained(
            pretrained_path, padding=False
        )

        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:

                smiles_dataset = SMILESTokenizerDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    smiles_language=smiles_language,
                    iterate_dataset=False,
                )

                config_file = os.path.join(pretrained_path, TOKENIZER_CONFIG_FILE)
                with open(config_file, encoding="utf-8") as fp:
                    max_length = json.load(fp)['max_token_sequence_length']
            c_index = smiles_dataset.smiles_language.token_to_index['C']
            o_index = smiles_dataset.smiles_language.token_to_index['O']

            sample = 0
            self.assertListEqual(
                smiles_dataset[sample].numpy().flatten().tolist(),
                [c_index, c_index, o_index],
            )

    def test_kwargs_read_smi(self):
        with TestFileContent(
            os.linesep.join(
                [
                    'CHEMBL545	metadata	CCO	and	so	on',
                    'CHEMBL17564	metadata	C	and	so	on',
                    'CHEMBL14688	metadata	CO	and	so	on',
                    'CHEMBL602	metadata	NCCS	and	so	on',
                ]
            )
        ) as test_file:
            smiles_dataset = SMILESTokenizerDataset(
                test_file.filename,
                index_col=0,
                names=['METADATA', 'SMILES', 'AND', 'SO', 'ON'],
            )
            self.assertEqual(smiles_dataset.dataset[2], 'CO')


class TestSMILESTokenizerDatasetLazy(TestSMILESTokenizerDatasetEager):
    """Testing SMILES dataset with lazy backend."""

    def setUp(self):
        self.backend = 'lazy'
        print(f'backend is {self.backend}')
        self.content = CONTENT
        self.other_content = MORE_CONTENT
        self.longest = LONGEST


if __name__ == '__main__':
    unittest.main()



"""Testing basic ways to setup a dataset."""
import unittest
import uuid
from copy import copy

import numpy as np
import pandas as pd
from torch.utils.data import DataLoader

from pytoda.datasets import (
    ConcatKeyDataset,
    DatasetDelegator,
    KeyDataset,
    indexed,
    keyed,
)
from pytoda.types import Hashable


class Indexed(KeyDataset):
    """As DataFrameDataset but only implementing necessary methods."""

    def __init__(self, df):
        self.df = df
        super().__init__()

    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):
        return self.df.iloc[index].values

    def get_key(self, index: int) -> Hashable:
        """Get key from integer index."""
        return self.df.index[index]

    def get_index(self, key: Hashable) -> int:
        """Get index for first datum mapping to the given key."""
        # item will raise if not single value (deprecated in pandas)
        try:
            indexes = np.nonzero(self.df.index == key)[0]
            return indexes.item()
        except ValueError:
            if len(indexes) == 0:
                raise KeyError
            else:
                # key not unique, return first as ConcatKeyDataset
                return indexes[0]


class Delegating(DatasetDelegator):
    """NOT implementing methods (and only built-ins from inheritance)."""

    def __init__(self, data):
        self.dataset = data


class TestBaseDatasets(unittest.TestCase):
    """Testing dataset for base methods."""

    length = 11  # of a single dataset
    dims = 5

    def random_data(self, length, dims):
        an_array = np.random.randn(self.length, self.dims)
        keys = [str(uuid.uuid4()) for _ in range(self.length)]
        a_df = pd.DataFrame(an_array, index=keys)
        return keys, a_df, Indexed(a_df)

    def setUp(self):
        (self.a_1st_keys, self.a_1st_df, self.a_1st_ds) = self.random_data(
            self.length, self.dims
        )
        (self.a_2nd_keys, self.a_2nd_df, self.a_2nd_ds) = self.random_data(
            self.length, self.dims
        )

        self.delegating_ds = Delegating(self.a_1st_ds)

        # KeyDataset.__add__ i.e. ConcatKeyDataset
        self.concat_ds = self.delegating_ds + self.a_2nd_ds
        self.concat_keys = self.a_1st_keys + self.a_2nd_keys

    def assertListedEqual(self, listable1, listable2):
        """Easier comparison between arrays, series and or lists."""
        self.assertListEqual(list(listable1), list(listable2))

    def test_delegation_dir(self):
        # stacking delegation
        ds_dir = dir(Delegating(Delegating(self.delegating_ds)))
        # delegated to Indexed
        self.assertIn('get_key', ds_dir)
        self.assertIn('get_index', ds_dir)
        # delegated to KeyDataset
        self.assertIn('get_item_from_key', ds_dir)
        self.assertIn('keys', ds_dir)
        self.assertIn('has_duplicate_keys', ds_dir)
        # futile, as built-ins delegation needed hardcoding in DatasetDelegator
        self.assertIn('__len__', ds_dir)  # see test___len__
        self.assertIn('__getitem__', ds_dir)  # see test___getitem__
        self.assertIn('__add__', ds_dir)  # see tests on self.concat_ds

        # no tests on implementation specific attributes here

    def test___len__(self) -> None:
        """Test __len__."""
        self.assertEqual(len(self.a_1st_ds), self.length)
        self.assertEqual(len(self.a_2nd_ds), self.length)
        self.assertEqual(len(self.delegating_ds), self.length)
        self.assertEqual(len(self.concat_ds), 2 * self.length)

    def test___getitem__(self) -> None:
        """Test __getitem__."""

        i = 0
        self.assertListedEqual(self.a_1st_ds[i], self.a_1st_df.iloc[i])
        self.assertListedEqual(self.a_2nd_ds[i], self.a_2nd_df.iloc[i])
        self.assertListedEqual(self.delegating_ds[i], self.a_1st_df.iloc[i])
        # first in datasets
        self.assertListedEqual(self.concat_ds[i], self.a_1st_df.iloc[i])

        i = -1
        self.assertListedEqual(self.a_1st_ds[i], self.a_1st_df.iloc[i])
        self.assertListedEqual(self.a_2nd_ds[i], self.a_2nd_df.iloc[i])
        self.assertListedEqual(self.delegating_ds[i], self.a_1st_df.iloc[i])
        # last in datasets
        self.assertListedEqual(self.concat_ds[i], self.a_2nd_df.iloc[i])

    def _test__getitem__modified(self, mutate_copy) -> None:
        """Test __getitem__ returning tuple with item first."""
        for i in [0, -1]:
            self.assertListedEqual(
                mutate_copy(self.a_1st_ds)[i][0], self.a_1st_df.iloc[i]
            )
            self.assertListedEqual(
                mutate_copy(self.a_2nd_ds)[i][0], self.a_2nd_df.iloc[i]
            )
            self.assertListedEqual(
                mutate_copy(self.delegating_ds)[i][0], self.a_1st_df.iloc[i]
            )

        # first in datasets
        self.assertListedEqual(mutate_copy(self.concat_ds)[0][0], self.a_1st_df.iloc[0])
        # last in datasets
        self.assertListedEqual(
            mutate_copy(self.concat_ds)[-1][0], self.a_2nd_df.iloc[-1]
        )

    def test__getitem__mutating_utils(self):
        self._test__getitem__modified(mutate_copy=indexed)
        self._test__getitem__modified(mutate_copy=keyed)

    def test_data_loader(self) -> None:
        """Test data_loader."""

        batch_size = 4
        a_1st_dl = DataLoader(self.a_1st_ds, batch_size=batch_size, shuffle=True)
        full_batches = self.length // batch_size

        for batch_index, batch in enumerate(a_1st_dl):
            if batch_index >= full_batches:  # if drop_last
                self.assertEqual(batch.shape, (self.length % batch_size, self.dims))
            else:
                self.assertEqual(batch.shape, (batch_size, self.dims))

        # concatenated
        concat_dl = DataLoader(self.concat_ds, batch_size=batch_size, shuffle=True)
        full_batches = (2 * self.length) // batch_size

        for batch_index, batch in enumerate(concat_dl):
            if batch_index >= full_batches:  # if drop_last
                self.assertEqual(
                    batch.shape, ((2 * self.length) % batch_size, self.dims)
                )
            else:
                self.assertEqual(batch.shape, (batch_size, self.dims))

    def _test_item_independent(self, ds, keys, index):
        key = keys[index]
        positive_index = index % len(ds)
        # get_key (support for negative index?)
        self.assertEqual(key, ds.get_key(positive_index))
        self.assertEqual(key, ds.get_key(index))
        # get_index
        self.assertEqual(positive_index, ds.get_index(key))
        # keys
        self.assertListedEqual(keys, ds.keys())
        # duplicate keys
        self.assertFalse(ds.has_duplicate_keys)

    def _test_base_methods(self, ds, keys, index):
        key = keys[index]
        self._test_item_independent(ds, keys, index)
        # get_item_from_key
        self.assertListedEqual(ds[index], ds.get_item_from_key(key))
        # in case of returning a tuple:
        # for from_index, from_key in zip(ds[index], ds.get_item_from_key(key)):  # noqa
        #     self.assertListedEqual(from_index, from_key)

    def _test_keyed_util(self, ds, keys, index):
        ds_ = keyed(ds)
        key = keys[index]
        self._test_item_independent(ds_, keys, index)

        # modified methods
        item_of_i, k_of_i = ds_[index]
        item_of_k, k_of_k = ds_.get_item_from_key(key)

        # get_item_from_key
        self.assertListedEqual(item_of_i, item_of_k)
        self.assertTrue(key == k_of_i and key == k_of_k)

    def _test_indexed_util(self, ds, keys, index):
        ds_ = indexed(ds)
        key = keys[index]
        positive_index = index % len(ds_)
        self._test_item_independent(ds_, keys, index)

        # modified methods
        item_of_i, i_of_i = ds_[index]
        item_of_k, i_of_k = ds_.get_item_from_key(key)

        # get_item_from_key
        self.assertListedEqual(item_of_i, item_of_k)
        self.assertTrue(positive_index == i_of_i and positive_index == i_of_k)

    def _test_stacked_indexed_keyed_util(self, ds, keys, index):
        ds_ = indexed(keyed(indexed(ds)))
        key = keys[index]
        positive_index = index % len(ds_)
        self._test_item_independent(ds_, keys, index)

        # modified methods
        (((item_of_i, i_of_i0), k_of_i), i_of_i1) = ds_[index]
        (((item_of_k, i_of_k0), k_of_k), i_of_k1) = ds_.get_item_from_key(key)

        # get_item_from_key
        self.assertListedEqual(item_of_i, item_of_k)
        self.assertTrue(key == k_of_i and key == k_of_k)
        self.assertTrue(
            positive_index == i_of_i0
            and positive_index == i_of_i1
            and positive_index == i_of_k0
            and positive_index == i_of_k1
        )

    def test_all_base_for_indexed_methods_and_copy(self):
        (other_keys, _, other_ds) = self.random_data(self.length, self.dims)

        for ds, keys in [
            (self.a_1st_ds, self.a_1st_keys),
            (self.a_2nd_ds, self.a_2nd_keys),
            (self.delegating_ds, self.a_1st_keys),
            (self.concat_ds, self.concat_keys),
            # test shallow copy (not trivial with delegation)
            (copy(self.a_1st_ds), self.a_1st_keys),
            (copy(self.a_2nd_ds), self.a_2nd_keys),
            (copy(self.delegating_ds), self.a_1st_keys),
            (copy(self.concat_ds), self.concat_keys),
        ]:
            index = -1
            self._test_indexed_util(ds, keys, index)
            self._test_keyed_util(ds, keys, index)
            self._test_stacked_indexed_keyed_util(ds, keys, index)
            self._test_base_methods(ds, keys, index)

            # again with self delegation and concatenation
            ds = ConcatKeyDataset([Delegating(other_ds), Delegating(ds)])
            index = self.length + 1  # dataset_index == 1
            keys = other_keys + keys
            self._test_indexed_util(ds, keys, index)
            self._test_keyed_util(ds, keys, index)
            self._test_stacked_indexed_keyed_util(ds, keys, index)
            self._test_base_methods(ds, keys, index)
            # get_index_pair
            self.assertTupleEqual((1, index - self.length), ds.get_index_pair(index))
            # get_key_pair
            self.assertTupleEqual((1, keys[index]), ds.get_key_pair(index))
            # ConcatKeyDataset is not a DatasetDelegator
            self.assertNotIn('df', dir(ds))

        index == self.length + 1
        duplicate_ds = other_ds + other_ds
        self.assertTrue(duplicate_ds.has_duplicate_keys)
        # duplicate keys lookup returns first in this case
        self.assertNotEqual(index, duplicate_ds.get_index(duplicate_ds.get_key(index)))


if __name__ == '__main__':
    unittest.main()



"""Implementation of _SmiLazyDataset."""
from ..files import read_smi
from ..types import Hashable, Sequence
from ._cache_datasource import _CacheDatasource
from .base_dataset import KeyDataset


class _SmiLazyDataset(KeyDataset, _CacheDatasource):
    """
    .smi dataset using lazy loading.

    Suggested when handling datasets that can't fit in the device memory.
    In case of datasets fitting in device memory consider using
    _SmiEagerDataset for better performance.
    """

    def __init__(
        self,
        smi_filepath: str,
        index_col: int = 1,
        chunk_size: int = 10000,
        name: str = 'SMILES',
        names: Sequence[str] = None,
    ) -> None:
        """
        Initialize a .smi dataset.

        Args:
            smi_filepath (str): path to .smi file.
            index_col (int): Data column used for indexing, defaults to 1.
            chunk_size (int): size of the chunks. Defaults to 10000.
            name (str): type of dataset, used to index columns in smi, and must
                be in names. Defaults to 'SMILES'.
            names (Sequence[str]): User-assigned names given to the columns.
                Defaults to `[name]`.
        """
        _CacheDatasource.__init__(self, fit_size_limit_filepath=smi_filepath)
        KeyDataset.__init__(self)
        self.smi_filepath = smi_filepath
        self.name = name
        self.names = names or [name]
        self.index_col = index_col
        self.chunk_size = chunk_size
        self.key_to_index_mapping = {}
        index = 0
        self.ordered_keys = []
        for chunk in read_smi(
            self.smi_filepath,
            index_col=self.index_col,
            chunk_size=self.chunk_size,
            names=self.names,
        ):
            for key, row in chunk.iterrows():
                self.cache[index] = row[self.name]
                self.key_to_index_mapping[key] = index
                index += 1
                self.ordered_keys.append(key)

        self.number_of_samples = len(self.ordered_keys)

    def __len__(self) -> int:
        """Total number of samples."""
        return self.number_of_samples

    def __getitem__(self, index: int) -> str:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            str: a SMILES for the current sample.
        """
        return self.cache[index]

    def get_key(self, index: int) -> Hashable:
        """Get key from integer index."""
        return self.ordered_keys[index]

    def get_index(self, key: Hashable) -> int:
        """Get index for first datum mapping to the given key."""
        return self.key_to_index_mapping[key]

    def keys(self):
        return iter(self.ordered_keys)

    @property
    def has_duplicate_keys(self):
        return self.number_of_samples != len(self.key_to_index_mapping)

    def __del__(self):
        """Delete the _SmiLazyDataset."""
        _CacheDatasource.__del__(self)



"""Implementation of _CsvEagerDataset."""
from collections import OrderedDict

import pandas as pd

from ..types import CallableOnSource, FeatureList, Optional
from ._csv_statistics import _CsvStatistics
from .dataframe_dataset import DataFrameDataset


class _CsvEagerDataset(DataFrameDataset, _CsvStatistics):
    """
    .csv dataset using eager loading.

    Suggested when handling datasets that can fit in the device memory.
    In case of out of memory errors consider using _CsvLazyDataset.
    """

    def __init__(
        self, filepath: str, feature_list: FeatureList = None, **kwargs
    ) -> None:
        """
        Initialize a .csv dataset.

        Args:
            filepath (str): path to .csv file.
            feature_list (FeatureList): a list of features. Defaults to None.
            kwargs (dict): additional parameters for pd.read_csv.
        """
        _CsvStatistics.__init__(
            self, filepath, feature_list=feature_list, **kwargs
        )  # calls setup_datasource

    def setup_datasource(self):
        """
        Setup the datasource and compute statistics.

        The dataframe is read, calling `self.preprocess_df` on it and setting
        up the data as source, collecting statistics of the data.

        NOTE:
        To read item with a different subset and order of features,
        use the function returned by `get_feature_fn` or use the
        feature_mapping to ensure integer indexing.

        Sets:
        feature_list (FeatureList): feature names in this datasource.
        feature_mapping (pd.Series): maps feature name to index in items.
        feature_fn (CallableOnSource): function that indexes datasource with
            the feature_list.
        """
        df = self.preprocess_df(pd.read_csv(self.filepath, **self.kwargs))
        # KeyDataset implementation, sets self.df
        DataFrameDataset.__init__(self, df)
        self.min_max_scaler.fit(self.df.values)
        self.standardizer.fit(self.df.values)
        self.notna_count = self.df.notna().sum().values

        self.feature_list = self.df.columns.tolist()
        self.feature_mapping = pd.Series(
            OrderedDict(
                [(feature, index) for index, feature in enumerate(self.feature_list)]
            )
        )
        self.feature_fn = self.get_feature_fn(self.feature_list)

    def get_feature_fn(self, feature_list: FeatureList) -> CallableOnSource:
        """Provides datasource specific indexing.

        Args:
            feature_list (FeatureList): subset of features to return in order.

        Returns:
            CallableOnSource: function that indexes datasource with the
                feature_list.
        """

        def feature_fn(df: pd.DataFrame) -> pd.DataFrame:
            return df[feature_list]

        return feature_fn

    def transform_datasource(
        self,
        transform_fn: CallableOnSource,
        feature_fn: CallableOnSource,
        impute: Optional[float] = None,
    ) -> None:
        """Apply scaling to the datasource.

        Args:
            transform_fn (CallableOnSource): transformation on source data.
            feature_fn (CallableOnSource): function that indexes datasource.
            impute (Optional[float]): NaN imputation with value if
                given. Defaults to None.
        """
        self.df = transform_fn(feature_fn(self.df))
        if impute is not None:
            self.df = self.df.fillna(impute)



"""Utils for the dataset module."""
from copy import copy

from ...types import Any, Files, Hashable, List, Tuple
from ..base_dataset import AnyBaseDataset, ConcatKeyDataset
from .factories import BACKGROUND_TENSOR_FACTORY


def sizeof_fmt(num, suffix='B'):
    """
    Human readable file size. Source: https://stackoverflow.com/a/1094933
    """
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def concatenate_file_based_datasets(
    filepaths: Files, dataset_class: type, **kwargs
) -> ConcatKeyDataset:
    r"""
    Concatenate file-based datasets into a single one, with the ability to
        get the source dataset of items.

    Args:
        filepaths (Files): list of filepaths.
        dataset_class (type): dataset class reading from file.
            Supports KeyDataset and DatasetDelegator.
            For pure torch.utils.data.Dataset the returned instance can
            still be used like a `pytoda.datasets.TransparentConcatDataset`,
            but methods depending on key lookup will fail.
        kwargs (dict): additional args for dataset_class.__init__(filepath, \**kwargs).

    Returns:
        ConcatKeyDataset: the concatenated dataset.
    """

    return ConcatKeyDataset(
        datasets=[dataset_class(filepath, **kwargs) for filepath in filepaths]
    )


def indexed(dataset: AnyBaseDataset) -> AnyBaseDataset:
    """
    Returns mutated shallow copy of passed dataset instance, where indexing
    behavior is changed to additionally returning index.
    """
    default_getitem = dataset.__getitem__  # bound method
    default_from_key = dataset.get_item_from_key  # bound method

    def return_item_index_tuple(self, index: int) -> Tuple[Any, int]:
        return_index = len(self) + index if index < 0 else index
        return default_getitem(index), return_index

    def return_item_index_tuple_from_key(self, key: Hashable) -> Tuple[Any, int]:
        """prevents `get_item_from_key` to call new indexed __getitem__"""
        return default_from_key(key), dataset.get_index(key)

    methods = {
        '__getitem__': return_item_index_tuple,
        'get_item_from_key': return_item_index_tuple_from_key,
    }
    ds = copy(dataset)
    ds.__class__ = type(
        f'Indexed{type(dataset).__name__}', (dataset.__class__,), methods
    )
    return ds


def keyed(dataset: AnyBaseDataset) -> AnyBaseDataset:
    """
    Returns mutated shallow copy of passed dataset instance, where indexing
    behavior is changed to additionally returning key.
    """
    default_getitem = dataset.__getitem__  # bound method
    default_from_key = dataset.get_item_from_key  # bound method

    def return_item_key_tuple(self, index: int) -> Tuple[Any, Hashable]:
        return (default_getitem(index), dataset.get_key(index))

    def return_item_key_tuple_from_key(self, key: Hashable) -> Tuple[Any, Hashable]:
        """prevents `get_item_from_key` to call new keyed __getitem__"""
        return default_from_key(key), key

    methods = {
        '__getitem__': return_item_key_tuple,
        'get_item_from_key': return_item_key_tuple_from_key,
    }
    ds = copy(dataset)
    ds.__class__ = type(f'Keyed{type(dataset).__name__}', (dataset.__class__,), methods)
    return ds


def pad_item(
    item: Tuple,
    padding_modes: List[str],
    padding_values: List,
    max_length: int,
) -> Tuple:
    """Padding function for a single item of a batch.

    Args:
        item (Tuple): Tuple returned by the __getitem__ function of a Dataset class.
        padding_modes (List[str]): The type of padding to perform for each datum in item.
            Options are 'constant' for constant value padding, and 'range' to fill the
            tensor with a range of values.
        padding_values (List): The values with which to fill the background tensor for padding.
            Can be a constant value or a range depending on the datum to pad in item.
        max_length (int): The maximum length to which the datum should be padded.

    Returns:
        Tuple: Tuple of tensors padded according to the given specifications.

    NOTE: pad_item function uses trailing dimensions as the repetitions argument
          for range_tensor(), since the 'length' of the set is covered by the
          value_range. That is, if a tensor of shape (5,) is required for
          padding_mode 'range' then () is passed as shape into range_tensor
          function which will repeat range(5) exactly once thus giving us a (5,) tensor.
    """
    # for each Tensor in the list we determine the output dimensions
    max_sizes = [datum.size() for datum in item]
    out_dimses = [
        (max_length, *max_sizes[i][1:])
        if padding_modes[i] == 'constant'
        else (*max_sizes[i][1:],)
        for i in range(len(max_sizes))
    ]

    out_tensors = [
        BACKGROUND_TENSOR_FACTORY[mode](value, out_dims)
        for out_dims, mode, value in zip(out_dimses, padding_modes, padding_values)
    ]

    for datum_index, tensor in enumerate(item):
        length = tensor.size(0)
        out_tensors[datum_index][:length, ...] = tensor
    return out_tensors



"""SMILES transforms."""
import logging
import re
from copy import deepcopy

import rdkit  # Needs import before torch in some envs
import numpy as np
import torch
from rdkit import Chem
from rdkit.Chem import AllChem
from selfies import encoder as selfies_encoder

import pytoda
from pytoda.warnings import device_warning

from ..transforms import Compose, LeftPadding, Randomize, StartStop, ToTensor, Transform
from ..types import Indexes, Tensor, Union

logger = logging.getLogger('pytoda_SMILES_transforms')


def compose_smiles_transforms(
    canonical: bool = False,
    augment: bool = False,
    kekulize: bool = False,
    all_bonds_explicit: bool = False,
    all_hs_explicit: bool = False,
    remove_bonddir: bool = False,
    remove_chirality: bool = False,
    selfies: bool = False,
    sanitize: bool = True,
    device: torch.device = None,
) -> Compose:
    """Setup a composition of SMILES to SMILES (or SELFIES) transformations.

    Args:
        canonical (bool, optional): performs canonicalization of SMILES
            (one original string for one molecule). If True, then other
            transformations (augment etc, see below) do not apply. Defaults to
            False.
        augment (bool, optional): perform SMILES augmentation. Defaults to
            False.
        kekulize (bool, optional): kekulizes SMILES (implicit aromaticity
            only). Defaults to False.
        all_bonds_explicit (bool, optional): makes all bonds explicit.
            Defaults to False, only applies if kekulize is True.
        all_hs_explicit (bool, optional): makes all hydrogens explicit.
            Defaults to False, only applies if kekulize is True.
        remove_bonddir (bool, optional): remove directional info of bonds.
            Defaults to False.
        remove_chirality (bool, optional): remove chirality information.
            Defaults to False.
        selfies (bool, optional): whether selfies is used instead of
            smiles. Defaults to False.
        sanitize (bool, optional): RDKit sanitization of the molecule.
            Defaults to True.
        device (torch.device): DEPRECATED

    Returns:
        Compose: A Callable that applies composition of SMILES transforms.
    """

    device_warning(device)
    # Build up composition from optional SMILES to SMILES transformations
    smiles_transforms = []
    if canonical:
        smiles_transforms += [Canonicalization(sanitize=sanitize)]
    else:
        if remove_bonddir or remove_chirality:
            smiles_transforms += [
                RemoveIsomery(bonddir=remove_bonddir, chirality=remove_chirality)
            ]
        if kekulize:
            smiles_transforms += [
                Kekulize(
                    all_bonds_explicit=all_bonds_explicit,
                    all_hs_explicit=all_hs_explicit,
                    sanitize=sanitize,
                )
            ]
        elif all_bonds_explicit or all_hs_explicit or sanitize:
            smiles_transforms += [
                NotKekulize(
                    all_bonds_explicit=all_bonds_explicit,
                    all_hs_explicit=all_hs_explicit,
                    sanitize=sanitize,
                )
            ]
        if augment:
            smiles_transforms += [
                Augment(
                    kekule_smiles=kekulize,
                    all_bonds_explicit=all_bonds_explicit,
                    all_hs_explicit=all_hs_explicit,
                    sanitize=sanitize,
                )
            ]
        if selfies:
            smiles_transforms += [Selfies()]

    return Compose(smiles_transforms)


def compose_encoding_transforms(
    randomize: bool = False,
    add_start_and_stop: bool = False,
    start_index: int = 2,
    stop_index: int = 3,
    padding: bool = False,
    padding_length: int = None,
    padding_index: int = 0,
) -> Compose:
    """Setup a composition of token indexes to token indexes transformations.

    Args:
        randomize (bool, optional): perform a true randomization of
            token indexes. Defaults to False.
        add_start_and_stop (bool, optional): add start and stop token
            indexes. Defaults to False.
        start_index (int, optional): index of start token in vocabulary.
            Default to 2.
        stop_index (int, optional): index of stop token in vocabulary.
            Default to 3.
        padding (bool, optional): pad sequences to given padding_length.
            Defaults to True.
        padding_length (int, optional): manually sets number of applied
            paddings, applies only if padding is True. Defaults to None, but
            must be passed in case of padding.
        padding_index (int, optional): index of padding token in vocabulary.
            Default to 0.

    Returns:
        Compose: A Callable that applies composition of transforms on
            token indexes.

    Note:
        Transformations can change the number of tokens.
    """
    encoding_transforms = []

    if randomize:
        encoding_transforms += [Randomize()]
    if add_start_and_stop:
        encoding_transforms += [StartStop(start_index, stop_index)]

    if padding:
        encoding_transforms += [
            LeftPadding(padding_length=padding_length, padding_index=padding_index)
        ]

    encoding_transforms += [ToTensor()]
    return Compose(encoding_transforms)


class SMILESToTokenIndexes(Transform):
    """Transform SMILES to token indexes using SMILES language."""

    def __init__(self, smiles_language) -> None:
        """
        Initialize a SMILES to token indexes object.

        Args:
            smiles_language (SMILESLanguage): a SMILES language.
                NOTE: No typing used to prevent circular import.
        """
        self.smiles_language = smiles_language

    def __call__(self, smiles: str) -> Union[Indexes, Tensor]:
        """
        Apply the SMILES tokenization transformation

        Args:
            smiles (str): a SMILES representation.

        Returns:
            Indexes: indexes representation for the SMILES provided.
        """
        return self.smiles_language.smiles_to_token_indexes(smiles)


class RemoveIsomery(Transform):
    """Remove isomery (isotopic and chiral specifications) from SMILES"""

    def __init__(self, bonddir=True, chirality=True, sanitize=True) -> None:
        """
        Initialize isomery removal.

        Args:
            bonddir (bool): whether bond direction information should be
                removed or not (default: {True})
            chirality (bool): whether chirality information should be removed
                (default: {True}).
        """

        self.chirality_dict = {'[': '', ']': '', '@': '', 'H': ''}
        self.bond_dict = {'/': '', '\\': ''}
        self.charge = re.compile(r'\[\w+\@?\@?[\+\-]\d?\]')
        self.multichar_atom = re.compile(r'\[[0-9]?[A-Za-z][a-z]?\w?[2-8]?\]')
        self.bonddir = bonddir
        self.chirality = chirality
        self.sanitize = sanitize

        if not self.bonddir and not self.chirality:
            self._call_fn = lambda smiles: smiles
        elif self.bonddir and not self.chirality:
            self.updates = str.maketrans(self.bond_dict)
            self._call_fn = self._isomery_call_fn
        elif self.chirality and not self.bonddir:
            self.updates = str.maketrans(self.chirality_dict)
            self._call_fn = self._isomery_call_fn
        else:
            self._call_fn = lambda smiles: Chem.MolToSmiles(
                Chem.MolFromSmiles(smiles, sanitize=sanitize),
                isomericSmiles=False,
                canonical=False,
            )

    def _isomery_call_fn(self, smiles: str) -> str:
        """
        Remove the stereoinfo of the SMILES. That can either be the removal of
        only the chirality information (at tetrahedral carbons with four
        different substituents) and/or the removal of the bond direction
        information (either the highest CIP-rated substituents on the same side
        of the two carbons of the double bond (Z) or on differnet sides (E)).

        NOTE: If bonddir=True and chirality=True, no transformation is applied.

        Args:
            smiles (str): a SMILES representation.

        Returns:
            str: SMILES representation of original smiles string with removed
                stereoinfo checked for validity
        """
        smiles = smiles.replace('[nH]', 'N')
        protect = []
        for m in self.charge.finditer(smiles):
            list_charg = list(range(m.start(), m.end()))
            protect += list_charg

        for m in self.multichar_atom.finditer(smiles):
            list_mc = list(range(m.start(), m.end()))
            protect += list_mc

        new_str = []
        for index, i in enumerate(smiles):
            new = i.translate(self.updates) if index not in protect else i
            new_str += list(new)
        smiles = ''.join(new_str).replace('N@@', 'N').replace('N@', 'N')

        try:
            Chem.SanitizeMol(Chem.MolFromSmiles(smiles, sanitize=self.sanitize))
            return smiles
        except TypeError:
            logger.warning(f'\nInvalid SMILES {smiles}')
            return ''

    def __call__(self, smiles: str) -> str:
        """
        Executable of RemoveIsomery class. The _call_fn is determined in
        the constructor based on the bonddir and chirality parameter.

        Args:
            smiles (str): a SMILES sequence.

        Returns:
            str: SMILES after the _call_fn was applied.
        """

        return self._call_fn(smiles)


class Kekulize(Transform):
    """Transform SMILES to Kekule version."""

    def __init__(self, all_bonds_explicit=False, all_hs_explicit=False, sanitize=True):

        # NOTE: Explicit bonds or Hs without Kekulization is not supported
        self.all_bonds_explicit = all_bonds_explicit
        self.all_hs_explicit = all_hs_explicit
        self.sanitize = sanitize

    def __call__(self, smiles: str) -> str:
        """
        Apply the kekulization transform.

        Args:
            smiles (str): a SMILES representation.
            all_bonds_explicit (bool): whether bonds are explicitly encoded.

        Returns:
            str: Kekulized SMILES of same molecule.
        """
        try:
            molecule = Chem.MolFromSmiles(smiles, sanitize=self.sanitize)
            if not self.sanitize:
                # Properties as valence are not calculated if sanitize is False
                molecule.UpdatePropertyCache(strict=False)
            Chem.Kekulize(molecule)
            return Chem.MolToSmiles(
                molecule,
                kekuleSmiles=True,
                allBondsExplicit=self.all_bonds_explicit,
                allHsExplicit=self.all_hs_explicit,
                canonical=False,
            )
        except Exception:
            logger.warning(
                f'\nInvalid SMILES {smiles}, no kekulization done '
                f'bondsExplicit: {self.all_bonds_explicit} & HsExplicit: '
                f'{self.all_hs_explicit} are also ignored.'
            )
            return smiles


class NotKekulize(Transform):
    """Transform SMILES without explicitly converting to Kekule version"""

    def __init__(self, all_bonds_explicit=False, all_hs_explicit=False, sanitize=True):
        self.all_bonds_explicit = all_bonds_explicit
        self.all_hs_explicit = all_hs_explicit
        self.sanitize = sanitize

    def __call__(self, smiles: str) -> str:
        """
        Apply transform.

        Args:
            smiles (str): a SMILES representation.

        Returns:
            str: SMILES of same molecule.
        """
        try:
            molecule = Chem.MolFromSmiles(smiles, sanitize=self.sanitize)
            return Chem.MolToSmiles(
                molecule,
                allBondsExplicit=self.all_bonds_explicit,
                allHsExplicit=self.all_hs_explicit,
                canonical=False,
            )
        except Exception:
            logger.warning(
                f'\nInvalid SMILES {smiles}, HsExplicit:{self.all_hs_explicit}'
                f'and bondsExplicit: {self.all_bonds_explicit} are ignored.'
            )
            return smiles


class Augment(Transform):
    """Augment a SMILES string, according to Bjerrum (2017)."""

    def __init__(
        self,
        kekule_smiles: bool = False,
        all_bonds_explicit: bool = False,
        all_hs_explicit: bool = False,
        sanitize: bool = True,
        seed: int = -1,
    ) -> None:
        """NOTE:  These parameter need to be passed down to the enumerator."""

        self.kekule_smiles = kekule_smiles
        self.all_bonds_explicit = all_bonds_explicit
        self.all_hs_explicit = all_hs_explicit
        self.sanitize = sanitize
        self.seed = seed
        if self.seed > -1:
            np.random.seed(self.seed)

    def __call__(self, smiles: str) -> str:
        """
        Apply the transform.

        Args:
            smiles (str): a SMILES representation.

        Returns:
            str: randomized SMILES representation.
        """
        molecule = Chem.MolFromSmiles(smiles, sanitize=self.sanitize)
        if molecule is None:
            logger.warning(f'\nAugmentation skipped for invalid mol: {smiles}')
            return smiles
        if not self.sanitize:
            molecule.UpdatePropertyCache(strict=False)
        atom_indexes = list(range(molecule.GetNumAtoms()))
        if len(atom_indexes) == 0:  # RDkit error handling
            return smiles
        np.random.shuffle(atom_indexes)
        renumbered_molecule = Chem.RenumberAtoms(molecule, atom_indexes)
        if self.kekule_smiles:
            Chem.Kekulize(renumbered_molecule)

        return Chem.MolToSmiles(
            renumbered_molecule,
            canonical=False,
            kekuleSmiles=self.kekule_smiles,
            allBondsExplicit=self.all_bonds_explicit,
            allHsExplicit=self.all_hs_explicit,
        )


class AugmentTensor(Transform):
    """
    Augment a SMILES (represented as a Tensor) according to Bjerrum (2017).
    """

    def __init__(
        self,
        smiles_language,
        kekule_smiles=False,
        all_bonds_explicit=False,
        all_hs_explicit=False,
        sanitize=True,
    ) -> None:
        """NOTE:  These parameter need to be passed down to the enumerator."""
        self.smiles_language = deepcopy(smiles_language)
        # Remove all SMILES transforms
        self.smiles_language.set_smiles_transforms(*[False] * 9)
        assert self.smiles_language.transform_smiles == Compose([])

        self.kekule_smiles = kekule_smiles
        self.all_bonds_explicit = all_bonds_explicit
        self.all_hs_explicit = all_hs_explicit
        self.sanitize = sanitize

    def update_smiles_language(self, smiles_language):
        if not isinstance(
            smiles_language, pytoda.smiles.smiles_language.SMILESLanguage
        ):
            raise ValueError('Please pass a SMILES language object')
        self.smiles_language = smiles_language

    def __call__(self, smiles_numerical: Union[Indexes, Tensor]) -> str:
        """
        Apply the transform.

        Args:
            smiles_numerical (Union[list, torch.Tensor]): either a SMILES
                represented as list of ints or a Tensor.

        Returns:
            torch.Tensor: randomized SMILES representation.
        """
        if type(smiles_numerical) == torch.Tensor:
            if smiles_numerical.ndim == 2:
                return self.__call__tensor(smiles_numerical)
            elif smiles_numerical.ndim == 1:
                smiles_numerical = smiles_numerical.cpu().numpy().flatten().tolist()
        if type(smiles_numerical) == list:
            smiles = self.smiles_language.token_indexes_to_smiles(smiles_numerical)
            try:
                molecule = Chem.MolFromSmiles(smiles, sanitize=self.sanitize)
                atom_indexes = list(range(molecule.GetNumAtoms()))
                if len(atom_indexes) == 0:  # RDkit error handling
                    return smiles
                np.random.shuffle(atom_indexes)
                renumbered_molecule = Chem.RenumberAtoms(molecule, atom_indexes)
                if self.kekule_smiles:
                    Chem.Kekulize(renumbered_molecule)

                augmented_smiles = Chem.MolToSmiles(
                    renumbered_molecule,
                    canonical=False,
                    kekuleSmiles=self.kekule_smiles,
                    allBondsExplicit=self.all_bonds_explicit,
                    allHsExplicit=self.all_hs_explicit,
                )
            except Exception:
                logger.warning(f'\nAugmentation skipped, mol invalid: {smiles}')
                augmented_smiles = smiles
            return self.smiles_language.smiles_to_token_indexes(augmented_smiles)

        raise TypeError('Please pass either a torch.Tensor of ndim 1, 2 or alist.')

    def __call__tensor(self, smiles_numerical: Tensor) -> torch.Tensor:
        """
        Wrapper of the transform for torch.Tensor.

        Args:
            smiles_numerical (torch.Tensor): a Tensor with SMILES represented
                as ints. Needs to have shape batch_size x sequence_length.
        Returns:
            torch.Tensor: Augmented SMILES representation of same shape.
        """
        # Infer the padding type to ensure returning tensor of same shape.
        if self.smiles_language.padding_index in smiles_numerical.flatten():

            padding = True
            left_padding = any(
                [
                    self.smiles_language.padding_index == row[0]
                    for row in smiles_numerical
                ]
            )
            right_padding = any(
                [
                    self.smiles_language.padding_index == row[-1]
                    for row in smiles_numerical
                ]
            )
            if (left_padding and right_padding) or (
                not left_padding and not right_padding
            ):
                raise ValueError(
                    'Could not uniqely infer padding type. Leftpadding was '
                    f'{left_padding}, right_padding was {right_padding}.'
                )
        else:
            padding = False

        seq_len = smiles_numerical.shape[1]

        # Loop over tensor (SMILES by SMILES) and augment. Exclude augmentation
        # if it violates the padding
        augmented = []
        for smiles in smiles_numerical:

            lenx = seq_len + 1
            while lenx > seq_len:
                augmented_smiles = self.__call__(smiles)
                lenx = len(augmented_smiles)
            if padding:
                pl = seq_len - len(augmented_smiles)
                pad = (0, pl) if right_padding else (pl, 0)
                augmented_smiles = torch.nn.functional.pad(
                    augmented_smiles, pad, value=self.smiles_language.padding_index
                )

            augmented.append(torch.unsqueeze(augmented_smiles, 0))

        augmented = torch.cat(augmented, dim=0)
        return augmented


class Selfies(Transform):
    """Convert a molecule from SMILES to SELFIES."""

    def __call__(self, smiles: str) -> str:
        return selfies_encoder(smiles)


class Canonicalization(Transform):
    """Convert any SMILES to RDKit-canonical SMILES.
    Example:
        An example::

            smiles = 'CN2C(=O)N(C)C(=O)C1=C2N=CN1C'
            c = Canonicalization()
            c(smiles)

        Result is: 'Cn1c(=O)c2c(ncn2C)n(C)c1=O'

    """

    def __init__(self, sanitize: bool = True) -> None:
        """Initialize a canonicalizer

        Args:
            sanitize (bool, optional): Whether molecule is sanitized. Defaults to True.
        """
        self.sanitize = sanitize

    def __call__(self, smiles: str) -> str:
        """
        Forward function of canonicalization.

        Args:
            smiles (str): SMILES string for canonicalization.

        Returns:
            str: Canonicalized SMILES string.
        """
        try:
            canon = Chem.MolToSmiles(
                Chem.MolFromSmiles(smiles, sanitize=self.sanitize), canonical=True
            )
            return canon
        except Exception:
            logger.warning(f'\nInvalid SMILES {smiles}, no canonicalization done')
            return smiles


class SMILESToMorganFingerprints(Transform):
    """Get fingerprints starting from SMILES."""

    def __init__(
        self, radius: int = 2, bits: int = 512, chirality=True, sanitize=False
    ) -> None:
        """
        Initialize a SMILES to fingerprints object.

        Args:
            radius (int): radius of the fingerprints.
            bits (int): bits used to represent the fingerprints.
        """
        self.radius = radius
        self.bits = bits
        self.chirality = chirality
        self.sanitize = sanitize

    def __call__(self, smiles: str) -> np.array:
        """
        Apply the transform.

        Args:
            smiles (str): a SMILES representation.

        Returns:
            np.array: the fingerprints.
        """
        try:
            molecule = Chem.MolFromSmiles(smiles, sanitize=self.sanitize)
            if not self.sanitize:
                molecule.UpdatePropertyCache(strict=False)
                AllChem.FastFindRings(molecule)
            fingerprint = AllChem.GetMorganFingerprintAsBitVect(
                molecule, self.radius, nBits=self.bits, useChirality=self.chirality
            )
        except Exception:
            logger.warning(f'\nInvalid SMILES {smiles}')
            molecule = Chem.MolFromSmiles('')
            fingerprint = AllChem.GetMorganFingerprintAsBitVect(
                molecule, self.radius, nBits=self.bits
            )
        array = np.zeros((1,))
        rdkit.DataStructs.ConvertToNumpyArray(fingerprint, array)
        return array



"""Protein language handling."""
import logging
from typing import Iterator

import dill
from upfp import parse_fasta

from ..files import read_smi
from ..types import Indexes, Tokenizer, Tokens
from .processing import HUMAN_KINASE_ALIGNMENT_VOCAB, IUPAC_VOCAB, UNIREP_VOCAB

logger = logging.getLogger(__name__)


class ProteinLanguage(object):
    """
    ProteinLanguage class.

    ProteinLanguage handle Protein data defining the vocabulary and
    utilities to manipulate it.
    """

    unknown_token = '<UNK>'

    def __init__(
        self,
        name: str = 'protein-language',
        amino_acid_dict: str = 'iupac',
        tokenizer: Tokenizer = list,
        add_start_and_stop: bool = True,
    ) -> None:
        """
        Initialize Protein language.

        Args:
            name (str): name of the ProteinLanguage.
            amino_acid_dict (str): Tokenization regime for amino acid
                sequence. Defaults to 'iupac', alternative is 'unirep' or
                'human-kinase-alignment'.
            tokenizer (Tokenizer): This needs to be a function used to tokenize
                the amino acid sequences. The default is list which simply
                splits the sequence character-by-character.
            add_start_and_stop (bool): add <START> and <STOP> in the sequence,
                of tokens. Defaults to True.
        """
        self.name = name
        self.dict = amino_acid_dict
        self.add_start_and_stop = add_start_and_stop

        if self.dict == 'iupac':
            self.token_to_index = IUPAC_VOCAB
        elif self.dict == 'unirep':
            self.token_to_index = UNIREP_VOCAB
        elif self.dict == 'human-kinase-alignment':
            self.token_to_index = HUMAN_KINASE_ALIGNMENT_VOCAB
        else:
            raise ValueError(
                "Choose dict as 'iupac' or 'unirep' or 'human-kinase-alignment' "
                f"(given was {amino_acid_dict})."
            )
        self.tokenizer = tokenizer
        self.setup_dict()

    def setup_dict(self) -> None:
        """
        Setup the dictionary.

        """
        # Setup dictionary
        self.sequence_tokens = [
            index for token, index in self.token_to_index.items() if '<' not in token
        ]
        self.number_of_tokens = len(self.token_to_index)
        self.index_to_token = {
            index: token for token, index in self.token_to_index.items()
        }

        if self.add_start_and_stop:
            self.max_token_sequence_length = 2
            self._get_total_number_of_tokens_fn = lambda tokens: len(tokens) + 2
            self._finalize_token_indexes_fn = lambda token_indexes: (
                [self.token_to_index['<START>']]
                + token_indexes
                + [self.token_to_index['<STOP>']]
            )
        else:
            self.max_token_sequence_length = 0
            self._get_total_number_of_tokens_fn = len
            self._finalize_token_indexes_fn = lambda token_indexes: token_indexes

        self.padding_index = self.token_to_index['<PAD>']
        self.start_index = self.token_to_index['<START>']
        self.stop_index = self.token_to_index['<STOP>']

    def __len__(self) -> int:
        """Number of characters the language knows."""
        return self.number_of_tokens

    @staticmethod
    def load(filepath: str) -> 'ProteinLanguage':
        """
        Static method to load a ProteinLanguage object.

        Args:
            filepath (str): path to the file.

        Returns:
            ProteinLanguage: the loaded Protein language object.
        """
        try:
            with open(filepath, 'rb') as f:
                protein_language = dill.load(f)
        except TypeError:
            # Necessary to load python3.7 pickled objects with >=3.8
            # For details see: https://github.com/uqfoundation/dill/pull/406
            storage = dill._dill._reverse_typemap['CodeType']
            dill._dill._reverse_typemap['CodeType'] = dill._dill._create_code
            with open(filepath, 'rb') as f:
                protein_language = dill.load(f)
            dill._dill._reverse_typemap['CodeType'] = storage
        return protein_language

    @staticmethod
    def dump(protein_language: 'ProteinLanguage', filepath: str):
        """
        Static method to save a Protein_language object to disk.

        Args:
            protein_language (ProteinLanguage): a ProteinLanguage object.
            filepath (str): path where to dump the ProteinLanguage.
        """
        with open(filepath, 'wb') as f:
            dill.dump(protein_language, f)

    def save(self, filepath: str):
        """
        Instance method to save/dump Protein language object.

        Args:
            filepath (str): path where to save the ProteinLanguage.
        """
        ProteinLanguage.dump(self, filepath)

    def _update_max_token_sequence_length(self, tokens: Tokens) -> None:
        """
        Update the max token sequence length.
        Uses method possibly overloaded by transformation setup to assess the
        length of tokens after transformations prior to their application.
        For example this allows handling start and stop tokens.

        Args:
            tokens (Tokens): tokens considered.
        """
        total_number_of_tokens = self._get_total_number_of_tokens_fn(tokens)
        if total_number_of_tokens > self.max_token_sequence_length:
            self.max_token_sequence_length = total_number_of_tokens

    def add_file(
        self,
        filepath: str,
        file_type: str = '.smi',
        index_col: int = 1,
        chunk_size: int = 100000,
    ) -> None:
        """
        Add a set of protein sequences from a file.

        Args:
            filepath (str): path to the file.
            file_type (str): Type of file, from {'.smi', '.csv', '.fasta',
                '.fasta.gz'}. If '.csv' is selected, it is assumed to be tab-
                separated.
            chunk_size (int): number of rows to read in a chunk.
                Defaults to 100000. Does not apply for fasta files.
            index_col (int): Data column used for indexing, defaults to 1, does
                not apply to fasta files.
        """
        if file_type not in ['.csv', '.smi', '.fasta', '.fasta.gz']:
            raise ValueError(
                "Please provide file of type {'.smi', '.csv', '.fasta','.fasta.gz'}"
            )

        if file_type == '.csv' or file_type == '.smi':
            try:
                for chunk in read_smi(
                    filepath,
                    chunk_size=chunk_size,
                    index_col=index_col,
                    names=['Sequence'],
                ):
                    for sequence in chunk['Sequence']:
                        self.add_sequence(sequence)
            except Exception:
                raise KeyError(
                    ".smi file needs to have 2 columns, index needs to be in "
                    f"column ({index_col}), sequences in the other."
                )
        elif file_type == '.fasta':
            database = parse_fasta(filepath, gzipped=False)
            for item in database:
                self.add_sequence(item['sequence'])

    def add_sequence(self, sequence: str) -> None:
        """
        Add a amino acid sequence to the language.

        Args:
            sequence (str): a sequence of amino acids.
        """
        tokens = self.tokenizer(sequence)
        self._update_max_token_sequence_length(tokens)

    def sequence_to_token_indexes_generator(self, sequence: str) -> Iterator[int]:
        """
        Transform tokens into indexes using a generator

        Args:
            sequence (str): an AAS representations

        Yields:
            Generator[int]: The generator of token indexes.
        """
        for token in self.tokenizer(sequence):
            if token not in self.token_to_index:
                logger.error(
                    'Replacing unknown token %s with %r', token, self.unknown_token
                )
                token = self.unknown_token
            yield self.token_to_index[token]

    def sequence_to_token_indexes(self, sequence: str) -> Indexes:
        """
        Transform character-level amino acid sequence (AAS) into a sequence of
        token indexes.

        Args:
            sequence (str): an AAS representation.

        Returns:
            Indexes: indexes representation for the AAS provided.
        """
        return self._finalize_token_indexes_fn(
            list(self.sequence_to_token_indexes_generator(sequence))
        )

    def token_indexes_to_sequence(self, token_indexes: Indexes) -> str:
        """
        Transform a sequence of token indexes into amino acid sequence.

        Args:
            token_indexes (Indexes): a sequence of token indexes.

        Returns:
            str: an amino acid sequence representation.
        """
        return ''.join(
            [
                self.index_to_token.get(token_index, '')
                for token_index in token_indexes
                if token_index in self.sequence_tokens
            ]
        )

    @property
    def method(self) -> str:
        """A string denoting the language method"""
        return self.dict



import torch
from scipy.optimize import linear_sum_assignment
from torch import nn
from torch.random import fork_rng
from torch.utils.data import Dataset

from pytoda.datasets.utils.utils import pad_item
from pytoda.types import Optional, Tensor, Tuple
from pytoda.warnings import device_warning


def hungarian_assignment(
    set_reference: torch.Tensor,
    set_matching: torch.Tensor,
    cost_metric_function: nn.Module,
) -> Tuple:
    """Compute targets for one training sample.

    Args:
        set_reference (torch.Tensor): Tensor with elements of set_reference.
        set_matching (torch.Tensor): Tensor with elements of set_matching.
        cost_metric_function (nn.Module): Function wrapped as an nn.Module that
            computes the metric used in constructing the cost matrix.

    Returns:
        Tuple: Tuple containing hungarian matching indices of set1 vs set2 and
            set2 vs set1.
    """

    cost_matrix = cost_metric_function(set_reference, set_matching)

    matrix = torch.zeros_like(cost_matrix)
    rows, cols = linear_sum_assignment(cost_matrix.cpu().numpy())
    matrix[rows, cols] = 1
    idx_12 = torch.from_numpy(cols)
    idx_21 = torch.nonzero(matrix.t(), as_tuple=True)[1]

    return idx_12, idx_21


def get_subsampling_indexes(
    min_set_length: int, max_set_length: int, permutation: Tensor, shuffle=True
) -> Tuple[Tensor, Tensor]:
    """Return indexers to remove random elements of an item and it's permutation.

    Args:
        min_set_length (int): Minimum number of elements in the set.
        max_set_length (int): Maximum number of elements in the set.
        permutation (Tensor): Tensor of integers defining a permutation, that are
            indices of a range in arbitrary order.
        shuffle (bool): The first returned indexer also shuffles the elements.

    Returns:
        Tuple[Tensor, Tensor, int]:
            A Tensor of integers for indexing a subset of elements (shuffled or not).
            A Tensor of integers for indexing the same elements in a permuted item.
            Number of elements in the item (length).
    """

    length = torch.randint(min_set_length, max_set_length + 1, (1,))
    if min_set_length != max_set_length:
        indexes_reference = torch.randperm(max_set_length)[:length]
        if not shuffle:
            indexes_reference = indexes_reference.sort().values
    else:
        indexes_reference = torch.arange(max_set_length)

    # This identifies indexes in set_matching that correspond to indexes_reference.
    # By doing so, we ensure that in the 'permute' case, the correct elements are retained during random cropping.

    indexes_matching = torch.tensor(
        [index for index, value in enumerate(permutation) if value in indexes_reference]
    )

    return indexes_reference, indexes_matching, length


class SetMatchingDataset(Dataset):
    """Base class for set matching datasets."""

    dataset: Dataset

    def __init__(
        self,
        min_set_length: int,
        cost_metric_function: nn.Module,
        set_padding_value: float = 0.0,
        seed: Optional[int] = None,
        shuffle: Optional[bool] = True,
        device: torch.device = None,
    ):
        """Base Class for set matching datasets, that allows returning subsets of
            varying length controlled by the passed min set length.

        Args:
            min_set_length (int): Lower bound on number of elements in the returned sets.
                This should be equal to the maximum item length of the passed dataset
                if varying set lengths are not desired.
            cost_metric_function (nn.Module): Function wrapped as an nn.Module that
                computes the metric used in constructing the cost matrix.
            set_padding_value (float): The constant value with which to pad each set.
                Defaults to 0.0. NOTE: No padding is done if min_set_length = max_set_length.
            seed (Optional[int]): If passed, all samples are generated once with
                this seed (using a local RNG only). Defaults to None, where
                individual samples are generated when the DistributionalDataset is
                indexed (using the global RNG).
            shuffle (Optional[bool]): Whether the sets should be shuffled again before subsampling.
                Adds another layer of randomness. Defaults to True.
            device (torch.device): DEPRECATED


        NOTE:
            1) Requires child classes to set the `dataset` attribute.
            2) Batches generated by DataLoader will have batch size in the first dim.
        """

        device_warning(device)
        # test that dataset is set (by child class)
        len(self.dataset)
        self.max_set_length = len(self.dataset[0])
        # Setup
        self.seed = seed

        self.get_cost_matrix = cost_metric_function

        self.min_set_length = min_set_length

        self.pad = not (min_set_length == self.max_set_length)

        self.set_padding_value = set_padding_value

        self.shuffle = shuffle

    @property
    def permutation(self) -> Tensor:
        """Class attribute that defines the permutation to use in creating set_matching.

        Raises:
            NotImplementedError: Not implemented by the base class. Attribute overwritten
                by the child class.
        """
        raise NotImplementedError

    def get_matching_set(self, index: int, reference_set: Tensor) -> Tensor:
        """Gets the corresponding set to match to the reference set.

        Args:
            index (int): The index to be sampled.
            reference_set (Tensor): Tensor that represents samples of the reference set.

        Raises:
            NotImplementedError: Not implemented by the base class. Function overwritten
                by the child class.
        """
        raise NotImplementedError

    def __len__(self) -> int:
        """Gets length of dataset.

        Returns:
            int: Length of the dataset being sampled.
        """
        return len(self.dataset)

    def __getitem__(self, index: int) -> Tuple:
        """Generates one sample from the dataset.

        Args:
            index (int): The index to be sampled.

        Returns:
            Tuple : Tuple containing sampled set1, sampled set2, hungarian
                matching indices of set1 vs set2 and set2 vs set1.
        """

        set_reference = self.dataset[index]

        enable_fork_rng = self.seed is not None
        with fork_rng(enabled=enable_fork_rng):
            if enable_fork_rng:
                seed = self.seed + index
                torch.manual_seed(seed)
                torch.cuda.manual_seed(seed)
                torch.cuda.manual_seed_all(seed)

            self.permute_indices = self.permutation
            indexes_reference, indexes_matching, length = get_subsampling_indexes(
                self.min_set_length,
                self.max_set_length,
                self.permute_indices,
                shuffle=self.shuffle,
            )

        set_matching = self.get_matching_set(index, set_reference)

        cropped_set_reference = set_reference[indexes_reference, :]
        cropped_set_matching = set_matching[indexes_matching, :]

        targets_12, targets_21 = hungarian_assignment(
            cropped_set_reference,
            cropped_set_matching,
            cost_metric_function=self.get_cost_matrix,
        )

        if not self.pad:
            return (
                cropped_set_reference,
                cropped_set_matching,
                targets_12,
                targets_21,
                length,
            )
        else:
            return (
                *pad_item(
                    item=(
                        cropped_set_reference,
                        cropped_set_matching,
                        targets_12,
                        targets_21,
                    ),
                    padding_modes=['constant', 'constant', 'range', 'range'],
                    padding_values=[
                        self.set_padding_value,
                        self.set_padding_value,
                        range(self.max_set_length),
                        range(self.max_set_length),
                    ],
                    max_length=self.max_set_length,
                ),
                length,
            )


class PairedSetMatchingDataset(SetMatchingDataset):
    """Dataset class for the case where set to match is another random set."""

    def __init__(
        self,
        dataset: Dataset,
        dataset_to_match: Dataset,
        min_set_length: int,
        cost_metric_function: nn.Module,
        set_padding_value: float,
        seed: Optional[int] = None,
        shuffle: Optional[bool] = True,
        noise_std: float = 0.0,
    ):
        """Pairs of sets and their hungarian assignments, where the set to match
        is provided by a second dataset.

        Args:
            dataset (Dataset): Object containing torch.utils.data.Dataset or
                child classes that represents samples of set_reference.
            dataset_to_match (Dataset): Object containing torch.utils.data.Dataset
                or child classes that represents samples of set_matching.
            min_set_length (int): Lower bound on number of elements in the returned sets.
                This should be equal to the maximum item length of the passed dataset
                if varying set lengths are not desired.
            set_padding_value (float): The constant value with which to pad each set.
                Defaults to 0.0. NOTE: No padding if min_set_length = max_set_length.
            cost_metric_function (nn.Module): Function wrapped as an nn.Module that
                computes the metric used in constructing the cost matrix.
            seed (Optional[int]): If passed, all samples are generated once with
                this seed (using a local RNG only). Defaults to None, where
                individual samples are generated when the DistributionalDataset is
                indexed (using the global RNG).
            shuffle (Optional[bool]): Whether the sets should be shuffled again before subsampling.
                Adds another layer of randomness. Defaults to True.
            noise_std (float, optional): Standard deviation to use in generating
                noise from a normal distribution with mean 0. Deafults to 0.0.
                Dummy variable for this class for consistency purposes.
        """
        # Setup dataset
        self.dataset = dataset
        self.dataset_to_match = dataset_to_match

        super().__init__(
            min_set_length,
            cost_metric_function,
            set_padding_value,
            seed,
            shuffle,
        )
        self.dummy_permutation = torch.arange(self.max_set_length)

    @property
    def permutation(self) -> Tensor:
        """Class attribute that defines the permutation to use in creating set_matching.

        Returns:
            Tensor: A fixed tensor containing the range of max_set_length.
        """
        return self.dummy_permutation

    def get_matching_set(self, index: int, set_reference: Tensor) -> Tensor:
        """Gets the corresponding set to match to the reference set.

        Args:
            index (int): The index to be sampled.
            reference_set (Tensor): Tensor that represents samples of the reference set.

        Returns:
            Tensor: Tensor of the corresponding matching set.
        """
        return self.dataset_to_match[index]


class PermutedSetMatchingDataset(SetMatchingDataset):
    """Dataset class for the case where set to match is permuted."""

    def __init__(
        self,
        dataset: Dataset,
        min_set_length: int,
        cost_metric_function: nn.Module,
        set_padding_value: float,
        seed: Optional[int] = None,
        noise_std: float = 0.0,
        shuffle: Optional[bool] = True,
    ):
        """Pairs of sets and their hungarian assignments, where the set to match
        is a permutation of the given sets.

        Args:
            dataset (Dataset): Object containing torch.utils.data.Dataset or child
                classes that represents samples of set_reference.
            min_set_length (int): Lower bound on number of elements in the returned sets.
                This should be equal to the maximum item length of the passed dataset
                if varying set lengths are not desired.
            set_padding_value (float): The constant value with which to pad each set.
                Defaults to 0.0. NOTE: No padding is done if min_set_length = max_set_length.
            cost_metric_function (nn.Module): Function wrapped as an nn.Module that
                computes the metric used in constructing the cost matrix.
            seed (Optional[int]): If passed, all samples are generated once with
                this seed (using a local RNG only). Defaults to None, where
                individual samples are generated when the DistributionalDataset is
                indexed (using the global RNG).
            noise_std (float, optional): Standard deviation to use in generating noise from
                a normal distribution with mean 0. Deafults to 0.0.
            shuffle (Optional[bool]): Whether the sets should be shuffled again before subsampling.
                Adds another layer of randomness. Defaults to True.
        """
        # Setup dataset
        self.dataset = dataset
        loc_device = torch.tensor(0.0)
        noise_std = torch.tensor(noise_std)
        self.noise = torch.distributions.normal.Normal(loc=loc_device, scale=noise_std)
        super().__init__(
            # is user choice
            min_set_length,
            cost_metric_function,
            set_padding_value,
            seed,
            shuffle,
        )

    @property
    def permutation(self) -> Tensor:
        """Class attribute that defines the permutation to use in creating set_matching.

        Returns:
            Tensor: Tensor of a randomly generated permutation of length max_set_length.
        """
        return torch.randperm(self.max_set_length)

    def get_matching_set(self, index: int, set_reference: Tensor) -> Tensor:
        """Gets the corresponding set to match to the reference set.

        Args:
            index (int): The index to be sampled.
            reference_set (Tensor): Tensor that represents samples of the reference set.

        Returns:
            Tensor: Tensor of the permuted reference set with additive noise.
        """
        enable_fork_rng = self.seed is not None
        with fork_rng(enabled=enable_fork_rng):
            if enable_fork_rng:
                seed = self.seed + index
                torch.manual_seed(seed)
                torch.cuda.manual_seed(seed)
                torch.cuda.manual_seed_all(seed)
            additive_noise = self.noise.sample(set_reference.size())

        return set_reference[self.permute_indices, :] + additive_noise



"""Testing SMILES processing."""
import unittest

import selfies as sf

from pytoda.smiles.processing import (
    kmer_smiles_tokenizer,
    spe_smiles_tokenizer,
    split_selfies,
    tokenize_selfies,
    tokenize_smiles,
)
from pytoda.smiles.transforms import Selfies


class TestProcessing(unittest.TestCase):
    """Testing processing."""

    def test_tokenize_smiles(self) -> None:
        """Test tokenize_smiles."""
        for smiles, ground_truth in [
            ('c1cnoc1', ['c', '1', 'c', 'n', 'o', 'c', '1']),
            (
                '[O-][n+]1ccccc1S',
                ['[O-]', '[n+]', '1', 'c', 'c', 'c', 'c', 'c', '1', 'S'],
            ),
            (
                'c1snnc1-c1ccccn1',
                [
                    'c',
                    '1',
                    's',
                    'n',
                    'n',
                    'c',
                    '1',
                    '-',
                    'c',
                    '1',
                    'c',
                    'c',
                    'c',
                    'c',
                    'n',
                    '1',
                ],
            ),
        ]:
            self.assertListEqual(tokenize_smiles(smiles), ground_truth)

    def test_kmer_smiles_tokenizer(self) -> None:
        """Test kmer_smiles_tokenizer."""

        smiles = ['c1ccc(/C=C/[C@H](C)O)cc', '[O-][n+]1ccccc1S', 'c1snnc1-c1ccccn1']

        k = 1
        stride = 1

        for s in smiles:
            self.assertListEqual(
                tokenize_smiles(s), kmer_smiles_tokenizer(s, k=k, stride=stride)
            )

        k = 2
        stride = 2
        for s, gt in zip(
            smiles,
            [
                ['c1', 'cc', 'c(', '/C', '=C', '/[C@H]', '(C', ')O', ')c'],
                ['[O-][n+]', '1c', 'cc', 'cc', '1S'],
                ['c1', 'sn', 'nc', '1-', 'c1', 'cc', 'cc', 'n1'],
            ],
        ):
            self.assertListEqual(kmer_smiles_tokenizer(s, k=k, stride=stride), gt)

        k = 2
        stride = 1
        for s, gt in zip(
            smiles,
            [
                [
                    'c1',
                    '1c',
                    'cc',
                    'cc',
                    'c(',
                    '(/',
                    '/C',
                    'C=',
                    '=C',
                    'C/',
                    '/[C@H]',
                    '[C@H](',
                    '(C',
                    'C)',
                    ')O',
                    'O)',
                    ')c',
                    'cc',
                ],
                ['[O-][n+]', '[n+]1', '1c', 'cc', 'cc', 'cc', 'cc', 'c1', '1S'],
                [
                    'c1',
                    '1s',
                    'sn',
                    'nn',
                    'nc',
                    'c1',
                    '1-',
                    '-c',
                    'c1',
                    '1c',
                    'cc',
                    'cc',
                    'cc',
                    'cn',
                    'n1',
                ],
            ],
        ):
            self.assertListEqual(kmer_smiles_tokenizer(s, k=k, stride=stride), gt)

        k = 4
        stride = 4
        for s, gt in zip(
            smiles,
            [
                ['c1cc', 'c(/C', '=C/[C@H]', '(C)O'],
                ['[O-][n+]1c', 'cccc'],
                ['c1sn', 'nc1-', 'c1cc', 'ccn1'],
            ],
        ):
            self.assertListEqual(kmer_smiles_tokenizer(s, k=k, stride=stride), gt)

        k = 4
        stride = 2
        for s, gt in zip(
            smiles,
            [
                [
                    'c1cc',
                    'ccc(',
                    'c(/C',
                    '/C=C',
                    '=C/[C@H]',
                    '/[C@H](C',
                    '(C)O',
                    ')O)c',
                ],
                ['[O-][n+]1c', '1ccc', 'cccc', 'cc1S'],
                ['c1sn', 'snnc', 'nc1-', '1-c1', 'c1cc', 'cccc', 'ccn1'],
            ],
        ):
            self.assertListEqual(kmer_smiles_tokenizer(s, k=k, stride=stride), gt)

    def test_spe_smiles_tokenizer(self) -> None:
        """Test spe_smiles_tokenizer."""
        for smiles, ground_truth in [
            ('c1ccc(/C=C/[C@H](C)O)cc', ['c1ccc(', '/C=C/', '[C@H](C)', 'O)', 'cc']),
            ('[O-][n+]1ccccc1S', ['[O-]', '[n+]1', 'ccccc1', 'S']),
            ('c1snnc1-c1ccccn1', ['c1s', 'nnc1', '-', 'c1ccccn1']),
        ]:
            self.assertListEqual(spe_smiles_tokenizer(smiles), ground_truth)

    def test_selfies_split(self) -> None:
        """Test tokenization by selfies package has not changed."""
        benzene = 'c1ccccc1'
        encoded_selfies = sf.encoder(benzene)
        # '[c][c][c][c][c][c][Ring1][Branch1_1]' v0.2.4
        # '[C][=C][C][=C][C][=C][Ring1][Branch1_2]' v1.0.2 (no aromatic)

        # sf.split_selfies returns generator
        symbols_benzene = list(sf.split_selfies(encoded_selfies))
        # before selfies 2.0.0 the last token is [Branch1_2]
        self.assertListEqual(
            symbols_benzene,
            ['[C]', '[=C]', '[C]', '[=C]', '[C]', '[=C]', '[Ring1]', '[=Branch1]'],
        )

        for smiles, ground_truth in [
            (
                'c1cnoc1',
                # before selfies 2.0.0 the 2 last token are [Expl=Ring1] and [Branch1_1]
                ['[C]', '[C]', '[=N]', '[O]', '[C]', '[=Ring1]', '[Branch1]'],
            ),
            (
                '[O-][n+]1ccccc1S',
                # before selfies 2.0.0 it is: [O-expl], [N+expl] and [=Branch1_2]
                [
                    '[O-1]',
                    '[N+1]',
                    '[=C]',
                    '[C]',
                    '[=C]',
                    '[C]',
                    '[=C]',
                    '[Ring1]',
                    '[=Branch1]',
                    '[S]',
                ],
            ),
            (
                'c1snnc1-c1ccccn1',
                # before selfies 2.0.0 it is: [Expl=Ring1], [Branch1_1] and [Branch1_2]
                [
                    '[C]',
                    '[S]',
                    '[N]',
                    '[=N]',
                    '[C]',
                    '[=Ring1]',
                    '[Branch1]',
                    '[C]',
                    '[=C]',
                    '[C]',
                    '[=C]',
                    '[C]',
                    '[=N]',
                    '[Ring1]',
                    '[=Branch1]',
                ],
            ),
        ]:
            self.assertListEqual(
                # list wrapping version
                split_selfies(sf.encoder(smiles)),
                ground_truth,
            )

    def test_tokenize_selfies_match(self) -> None:
        """Test deprecated tokenize_selfies."""
        for smiles in ['c1cnoc1', '[O-][n+]1ccccc1S', 'c1snnc1-c1ccccn1']:
            transform = Selfies()
            selfies = transform(smiles)
            self.assertListEqual(
                tokenize_selfies(selfies), list(sf.split_selfies(selfies))
            )

    def test_package_tokenizer(self) -> None:
        """Test package tokenizer."""
        from pytoda.smiles import vocab

        self.assertEqual(len(vocab), 575)


if __name__ == '__main__':
    unittest.main()



"""Testing Crawlers."""
import unittest

from pytoda.preprocessing.crawlers import (  # query_pubchem,; remove_pubchem_smiles,
    get_smiles_from_pubchem,
    get_smiles_from_zinc,
)


class TestCrawlers(unittest.TestCase):
    """Testing Crawlsers."""

    def test_get_smiles_from_zinc(self) -> None:
        """Test get_smiles_from_zinc"""

        # # ZINC is down since quite some time, hence we skip these tests
        return True

        # Test text mode
        drug = 'Aspirin'
        ground_truth = 'CC(=O)Oc1ccccc1C(=O)O'
        smiles = get_smiles_from_zinc(drug)
        self.assertEqual(smiles, ground_truth)

        # Test ZINC ID mode
        zinc_id = 53
        ground_truth = 'CC(=O)Oc1ccccc1C(=O)O'
        smiles = get_smiles_from_zinc(zinc_id)
        self.assertEqual(smiles, ground_truth)

    def test_get_smiles_from_pubchem(self) -> None:
        """Test get_smiles_from_zinc"""

        for sanitize in [True, False]:

            # Test text mode
            ground_truth = 'C1=CC(=CC=C1/C=C/C(=O)C2=C(C=C(C=C2)O)O)O'
            for query, drug in zip(['name', 'cid'], ['isoliquiritigenin', 638278]):
                smiles = get_smiles_from_pubchem(
                    drug,
                    use_isomeric=True,
                    kekulize=True,
                    sanitize=sanitize,
                    query_type=query,
                )
            self.assertEqual(smiles, ground_truth)

            ground_truth = 'C1=CC(=CC=C1C=CC(=O)C2=C(C=C(C=C2)O)O)O'
            for query, drug in zip(['name', 'cid'], ['isoliquiritigenin', 638278]):
                smiles = get_smiles_from_pubchem(
                    drug,
                    use_isomeric=False,
                    kekulize=True,
                    sanitize=sanitize,
                    query_type=query,
                )
            # mac-os irreproducible stochastic failure on ubuntu
            self.assertIn(smiles, [ground_truth, ''])

            drug = 'isoliquiritigenin'
            if not sanitize:
                with self.assertRaises(ValueError):
                    get_smiles_from_pubchem(
                        drug, use_isomeric=True, kekulize=False, sanitize=sanitize
                    )

                    get_smiles_from_pubchem(
                        drug, use_isomeric=False, kekulize=False, sanitize=sanitize
                    )
            else:
                ground_truth = 'O=C(/C=C/c1ccc(O)cc1)c1ccc(O)cc1O'

                smiles = get_smiles_from_pubchem(
                    drug, use_isomeric=True, kekulize=False, sanitize=sanitize
                )
                self.assertEqual(smiles, ground_truth)

                ground_truth = 'O=C(C=Cc1ccc(O)cc1)c1ccc(O)cc1O'
                smiles = get_smiles_from_pubchem(
                    drug, use_isomeric=False, kekulize=False, sanitize=sanitize
                )
                self.assertEqual(smiles, ground_truth)

    def test_query_pubchem(self) -> None:
        """Test query_pubchem"""
        pass
        # Disabled due to bug in pubchem api
        # smiles_list = [
        #     'O1C=CC=NC(=O)C1=O',
        #     'CC(N)S(O)(=O)C(C)CC(C(C)C)c1cc(F)cc(F)c1',
        #     'Clc1ccccc2ccnc12',
        # ]
        # ground_truths = [(True, 67945516), (False, -2), (False, -1)]
        # for gt, smiles in zip(ground_truths, smiles_list):
        #     self.assertTupleEqual(query_pubchem(smiles), gt)

    def test_remove_pubchem_smiles(self) -> None:
        """Test remove_pubchem_smiles"""
        pass

        # Disabled due to bug in pubchem api
        # smiles_list = [
        #     'O1C=CC=NC(=O)C1=O',
        #     'CC(N)S(O)(=O)C(C)CC(C(C)C)c1cc(F)cc(F)c1',
        #     'Clc1ccccc2ccnc12',
        # ]
        # ground_truth = ['CC(N)S(O)(=O)C(C)CC(C(C)C)c1cc(F)cc(F)c1', 'Clc1ccccc2ccnc12']
        # self.assertListEqual(remove_pubchem_smiles(smiles_list), ground_truth)


if __name__ == '__main__':
    unittest.main()



"""Transform utilities."""
import logging
import random
from copy import deepcopy
from typing import Any, Dict

import numpy as np
import torch

from .types import Indexes, TransformList

logger = logging.getLogger('pytoda_transforms')


class Transform(object):
    """Transform abstract class."""

    def __call__(self, sample: Any) -> Any:
        """
        Apply the transformation.

        Args:
            sample (Any): a sample to transform.

        Returns:
            Any: the transformed sample.
        """
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        """Equality comparison of Transform objects. Two transform instances are
        identical if the transforms are identical

        Args:
            other (Compose): Compose object for comparison.

        Returns:
            bool: Whether objects are identical
        """
        return vars(self) == vars(other)


class StartStop(Transform):
    """Add start and stop token indexes at beginning and end of sequence."""

    def __init__(self, start_index: int, stop_index: int):
        """
        Initialize a left padding token indexes object.

        Args:
            start_index (int): index of start token in vocabulary.
            start_index (int): index of stop token in vocabulary.
        """
        self.start_index = start_index
        self.stop_index = stop_index

    def __call__(self, token_indexes: Indexes) -> Indexes:
        """
        Apply the transform.

        Args:
            token_indexes (Indexes): token indexes.

        Returns:
            Indexes: Indexes representation with start and stop added.
        """
        return [self.start_index] + token_indexes + [self.stop_index]


class LeftPadding(Transform):
    """Left pad token indexes."""

    def __init__(self, padding_length: int, padding_index: int) -> None:
        """
        Initialize a left padding token indexes object.

        Args:
            padding_length (int): length of the padding.
            padding_index (int): padding index.
        """
        self.padding_length = padding_length
        self.padding_index = padding_index

    def __call__(self, token_indexes: Indexes) -> Indexes:
        """
        Apply the transform.

        Args:
            token_indexes (Indexes): token indexes.

        Returns:
            Indexes: indexes representation with given `padding_length`.
                token_indexes is cut short or left padded with `padding_index`.
        """
        try:
            if self.padding_length < len(token_indexes):
                logger.warning(
                    f'\n{token_indexes} is longer than padding length '
                    f'({self.padding_length}). End of string will be stripped '
                    'off.'
                )
                return token_indexes[: self.padding_length]
            else:
                return (self.padding_length - len(token_indexes)) * [
                    self.padding_index
                ] + token_indexes
        except TypeError as e:
            if self.padding_length is None:
                raise TypeError(
                    'padding_length=None was given but integer is required.'
                )
            else:
                raise e


class ToTensor(Transform):
    """Transform token indexes to torch tensor."""

    def __init__(self, dtype: torch.dtype = torch.short) -> None:
        """
        Initialize a token indexes to tensor object.

        Args:
            dtype (torch.dtype): data type. Defaults to torch.short.
        """

        if not isinstance(dtype, torch.dtype):
            raise TypeError(f'Dtype must be torch.dtype not {type(dtype)}')
        self.dtype = dtype

    def __call__(self, token_indexes: Indexes) -> torch.Tensor:
        """
        Apply the transform. Convert token_indexes (e.g. list) to a torch Tensor.

        Args:
            token_indexes (Indexes): token indexes.

        Returns:
            torch.Tensor: tensor representation of the token indexes.
        """
        return torch.tensor(token_indexes, dtype=self.dtype).view(-1, 1).squeeze()


class ListToTensor(Transform):
    """
    2D Version of ToTensor.
    """

    def __init__(self, dtype: torch.dtype = torch.float) -> None:
        """
        Initialize a token indexes to tensor object.

        Args:

            dtype (torch.dtype): data type. Defaults to torch.float.
        """
        if not isinstance(dtype, torch.dtype):
            raise TypeError(f'Dtype must be torch.dtype not {type(dtype)}')
        self.dtype = dtype

    def __call__(self, token_indexes: Indexes) -> torch.Tensor:
        """
        Apply the transform (Transform a list of token indexes to torch tensor).
            2D version of ToTensor.

        Args:
            token_indexes (Indexes): token indexes.

        Returns:
            torch.Tensor: tensor representation of the token indexes.
        """
        return torch.tensor(token_indexes, dtype=self.dtype)


class Randomize(Transform):
    """Randomize a sequence all tokens."""

    def __call__(self, tokens: Indexes) -> Indexes:
        """
        Args:
            tokens (Indexes): indexes representation for the SMILES to be
                randomized.
        Returns:
           Indexes: shuffled indexes representation of the molecule

        NOTE: If this transform is used on SMILES, it must not be applied on
            the raw SMILES string, but on the tokenized, numerical vectors
            (i.e. after SMILESToTokenIndexes).
        """
        smiles_tokens = deepcopy(tokens)
        np.random.shuffle(smiles_tokens)
        return smiles_tokens


class AugmentByReversing(Transform):
    """Augment an sequence by (eventually) flipping order"""

    def __init__(self, p: float = 0.5) -> None:
        """
        AugmentByReversing constructor.

        Args:
            p (float): Probability that reverting occurs.

        """
        if not isinstance(p, float):
            raise TypeError(f'Please pass float, not {type(p)}.')
        self.p = np.clip(p, 0.0, 1.0)

    def __call__(self, sequence: str) -> str:
        """
        Apply the transform.

        Args:
            sequnce (str): a sequence representation.

        Returns:
            str: Either the sequence itself, or the revesed sequence.
        """
        return sequence[::-1] if random.random() < self.p else sequence


class Compose(Transform):
    """
    Composes several transforms together.

    From:
    https://pytorch.org/docs/stable/torchvision/transforms.html#torchvision.transforms.Compose.
    """

    def __init__(self, transforms: TransformList) -> None:
        """
        Initialize a compose transform object.

        Args:
            transforms (TransformList): a list of transforms.
        """
        self.transforms = transforms

    def __call__(self, sample: Any) -> Any:
        """
        Apply the composition of transforms.

        Args:
            sample (Any): a sample to transform.

        Returns:
            Any: the transformed sample.
        """
        for transform in self.transforms:
            sample = transform(sample)
        return sample

    def __repr__(self) -> str:
        """
        Represent the composition as a string.

        Returns:
            str: a string representing the composed transformation.
        """
        format_string = self.__class__.__name__ + '('
        for transform in self.transforms:
            format_string += '\n'
            format_string += '\t{}'.format(transform)
        format_string += '\n)'
        return format_string

    def __eq__(self, other: Transform) -> bool:
        """Equality comparison of Compose objects. Two compose instances are identical
        if the transforms are identical

        Args:
            other (Compose): Compose object for comparison.

        Returns:
            bool: Whether objects are identical
        """
        if len(self.transforms) != len(other.transforms):
            return False
        else:
            return all([a == b for a, b in zip(self.transforms, other.transforms)])


class ExtractFromDict(Transform):
    """Extracts a value from a dictionary based on a key set in the constructor."""

    def __init__(self, key: str) -> None:
        """
        Args:
            key: The key used to access the dictionary
        """
        self.key = key

    def __call__(self, sample_dict: Dict[str, str]) -> str:
        """
        Extract value from a str-str-dictionary.

        Args:
            sample_dict: A dictionary containing the key set in constructor.

        Returns:
            The extracted string.
        """
        assert self.key in sample_dict, f'Key {self.key} not found in sample_dict'
        return sample_dict[self.key]


class DiscardLowercase(Transform):
    """Discard lower-case letters (and non-letter characters) from a sequence."""

    def __init__(
        self,
    ) -> None:
        """ """
        pass

    def __call__(self, sequence: str) -> str:
        """
        Apply the transform.

        Args:
            sequence (str):

        Returns:
            str:
        """
        ans = ''.join([x for x in sequence if (x >= 'A') and (x <= 'Z')])
        return ans


class ToUpperCase(Transform):
    """Convert all characters to uppercase."""

    def __init__(
        self,
    ) -> None:
        pass

    def __call__(self, sequence: str) -> str:
        """
        Apply the transform.

        Args:
            sequence (str): A sequence.

        Returns:
            str: A sequence in uppercase.
        """
        ans = sequence.upper()
        return ans



"""Datasets for smiles and transformations of smiles."""
import logging

import torch

from pytoda.warnings import device_warning

from ..smiles.processing import split_selfies
from ..smiles.smiles_language import SMILESLanguage, SMILESTokenizer
from ._smi_eager_dataset import _SmiEagerDataset
from ._smi_lazy_dataset import _SmiLazyDataset
from .base_dataset import DatasetDelegator
from .utils import concatenate_file_based_datasets

logger = logging.getLogger(__name__)


SMILES_DATASET_IMPLEMENTATIONS = {  # get class and acceptable keywords
    'eager': (_SmiEagerDataset, {'index_col', 'names'}),
    'lazy': (_SmiLazyDataset, {'chunk_size', 'index_col', 'names'}),
}  # name cannot be passed


class SMILESDataset(DatasetDelegator):
    """Dataset of SMILES."""

    def __init__(
        self,
        *smi_filepaths: str,
        backend: str = 'eager',
        name: str = 'smiles-dataset',
        device: torch.device = None,
        **kwargs,
    ) -> None:
        """
        Initialize a SMILES dataset.

        Args:
            smi_filepaths (Files): paths to .smi files.
            name (str): name of the SMILESDataset.
            backend (str): memory management backend.
                Defaults to eager, prefer speed over memory consumption.
            device (torch.device): DEPRECATED
            kwargs (dict): additional arguments for dataset constructor.
        """
        device_warning(device)
        # Parse language object and data paths
        self.smi_filepaths = smi_filepaths
        self.backend = backend
        self.name = name

        dataset_class, valid_keys = SMILES_DATASET_IMPLEMENTATIONS[self.backend]
        self.kwargs = dict((k, v) for k, v in kwargs.items() if k in valid_keys)
        self.kwargs['name'] = 'SMILES'

        self.dataset = concatenate_file_based_datasets(
            filepaths=self.smi_filepaths, dataset_class=dataset_class, **self.kwargs
        )

        DatasetDelegator.__init__(self)  # delegate to self.dataset
        if self.has_duplicate_keys:
            raise KeyError('Please remove duplicates from your .smi file.')


class SMILESTokenizerDataset(DatasetDelegator):
    """Dataset of token indexes from SMILES."""

    def __init__(
        self,
        *smi_filepaths: str,
        smiles_language: SMILESLanguage = None,
        canonical: bool = False,
        augment: bool = False,
        kekulize: bool = False,
        all_bonds_explicit: bool = False,
        all_hs_explicit: bool = False,
        remove_bonddir: bool = False,
        remove_chirality: bool = False,
        selfies: bool = False,
        sanitize: bool = True,
        randomize: bool = False,
        add_start_and_stop: bool = False,
        padding: bool = True,
        padding_length: int = None,
        vocab_file: str = None,
        iterate_dataset: bool = True,
        backend: str = 'eager',
        device: torch.device = None,
        name: str = 'smiles-encoder-dataset',
        **kwargs,
    ) -> None:
        """
        Initialize a dataset providing token indexes from source SMILES.

        The datasets transformations on smiles and encodings can be adapted,
        depending on the smiles_language used (see SMILESTokenizer).

        Args:
            smi_filepaths (Files): paths to .smi files.
            smiles_language (SMILESLanguage): a smiles language that transforms
                and encodes SMILES to token indexes. Defaults to None, where
                a SMILESTokenizer is instantited with the following arguments.
            canonical (bool): performs canonicalization of SMILES (one
                original string for one molecule), if True, then other
                transformations (augment etc, see below) do not apply
            augment (bool): perform SMILES augmentation. Defaults to False.
            kekulize (bool): kekulizes SMILES (implicit aromaticity only).
                Defaults to False.
            all_bonds_explicit (bool): Makes all bonds explicit. Defaults to
                False, only applies if kekulize = True.
            all_hs_explicit (bool): Makes all hydrogens explicit. Defaults to
                False, only applies if kekulize = True.
            randomize (bool): perform a true randomization of SMILES tokens.
                Defaults to False.
            remove_bonddir (bool): Remove directional info of bonds.
                Defaults to False.
            remove_chirality (bool): Remove chirality information.
                Defaults to False.
            selfies (bool): Whether selfies is used instead of smiles, defaults
                to False.
            sanitize (bool): RDKit sanitization of the molecule.
                Defaults to True.
            add_start_and_stop (bool): add start and stop token indexes.
                Defaults to False.
            padding (bool): pad sequences to longest in the smiles language.
                Defaults to True.
            padding_length (int): padding to match manually set length,
                applies only if padding is True. Defaults to None.
            vocab_file (str): Optional .json to load vocabulary. Defaults to
                None.
            iterate_dataset (bool): whether to go through all SMILES in the
                dataset to extend/build vocab, find longest sequence, and
                checks the passed padding length if applicable. Defaults to
                True.
            backend (str): memory management backend.
                Defaults to eager, prefer speed over memory consumption.
            name (str): name of the SMILESTokenizerDataset.
            device (torch.device): DEPRECATED
            kwargs (dict): additional arguments for dataset constructor.

        """
        device_warning(device)
        self.name = name
        self.dataset = SMILESDataset(*smi_filepaths, backend=backend, **kwargs)
        DatasetDelegator.__init__(self)  # delegate to self.dataset

        if smiles_language is not None:
            self.smiles_language = smiles_language
            params = (
                "canonical, augment, kekulize, all_bonds_explicit, selfies, sanitize, "
                "all_hs_explicit, remove_bonddir, remove_chirality, randomize, "
                "add_start_and_stop, padding, padding_length"
            )
            logger.error(
                'Since you provided a smiles_language, the following parameters to this'
                f' class will be ignored: {params}.\nHere are the problems:'
            )
            mismatch = False
            for p in params.split(','):
                if eval(p.strip()) != eval(f'smiles_language.{p.strip()}'):
                    logger.error(
                        f'Provided arg {p.strip()}:{eval(p.strip())} does not match the '
                        f'smiles_language value: {eval(f"smiles_language.{p.strip()}")}'
                        ' NOTE: smiles_language value takes preference!!'
                    )
                    mismatch = True
            if not mismatch:
                logger.error('Looking great, no problems found!')
            else:
                logger.error(
                    'To get rid of this, adapt the smiles_language *offline*, feed it'
                    'ready for intended usage, and adapt the constructor args to be '
                    'identical with their equivalents in the language object'
                )

        else:
            language_kwargs = {}  # SMILES default
            if selfies:
                language_kwargs = dict(
                    name='selfies-language', smiles_tokenizer=split_selfies
                )
            self.smiles_language = SMILESTokenizer(
                **language_kwargs,
                canonical=canonical,
                augment=augment,
                kekulize=kekulize,
                all_bonds_explicit=all_bonds_explicit,
                all_hs_explicit=all_hs_explicit,
                randomize=randomize,
                remove_bonddir=remove_bonddir,
                remove_chirality=remove_chirality,
                selfies=selfies,
                sanitize=sanitize,
                add_start_and_stop=add_start_and_stop,
                padding=padding,
                padding_length=padding_length,
            )

        if vocab_file:
            self.smiles_language.load_vocabulary(vocab_file)

        if iterate_dataset:
            # uses the smiles transforms
            self.smiles_language.add_dataset(self.dataset)

        try:
            if (
                self.smiles_language.padding
                and self.smiles_language.padding_length is None
            ):
                try:
                    # max_sequence_token_length has to be set somehow
                    if smiles_language is not None or iterate_dataset:
                        self.smiles_language.set_max_padding()
                except AttributeError:
                    raise TypeError(
                        'Setting a maximum padding length requires a '
                        'smiles_language with `set_max_padding` method. See '
                        '`SMILESTokenizer`.'
                    )
        except AttributeError:
            # SmilesLanguage w/o padding support passed.
            pass

    def __getitem__(self, index: int) -> torch.Tensor:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            torch.Tensor: a torch tensor of token indexes,
                for the current sample.
        """
        return self.smiles_language.smiles_to_token_indexes(self.dataset[index])



import torch
import torch.nn as nn

from ...types import Tensor


class WrapperCDist(nn.Module):
    """Wrapper for torch.cdist module for easy argument passing."""

    def __init__(self, p: int = 2) -> None:
        """Constructor.

        Args:
            p (int, optional): p value for the p-norm distance to calculate
                between each vector pair. Defaults to 2.
        """

        super(WrapperCDist, self).__init__()
        self.p = p

    def forward(self, set1: Tensor, set2: Tensor) -> Tensor:
        """Computes the pairwise p-norms.

        Args:
            set1 (Tensor): Input tensor of shape
                [batch_size, length1, dim]
            set2 (Tensor): Input tensor of shape
                [batch_size, length2, dim]

        Returns:
            Tensor: Tensor of shape [batch_size, length1, length2]
                representing the pairwise distances.
        """
        return torch.cdist(set1, set2, self.p)


class WrapperKLDiv(nn.Module):
    """Wrapper for KL-Divergence for easy argument passing."""

    def __init__(self, reduction: str = 'mean') -> None:
        """Constructor.

        Args:
            reduction (str, optional): One of 'none','batchmean','sum', 'mean'.
                Defaults to 'mean'.
        """

        super(WrapperKLDiv, self).__init__()

        self.reduction = reduction

    def forward(self, set1: Tensor, set2: Tensor) -> Tensor:
        """Computes the KL-Divergence.

        Args:
            set1 (Tensor): Input tensor of arbitrary shape.
            set2 (Tensor): Tensor of the same shape as input.

        Returns:
            Tensor: Scalar by default. if reduction = 'none', then same
                shape as input.
        """

        return nn.functional.kl_div(set1, set2, reduction=self.reduction)



"""Abstract implementation of _CsvStatistics."""
import copy
from functools import reduce

import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from ..types import CallableOnSource, FeatureList, List, Optional, OrderedDict, Tuple


class _CsvStatistics:
    """.csv abstract setup for dataset statistics."""

    def __init__(
        self,
        filepath: str,
        feature_list: FeatureList = None,
        pandas_dtype=None,
        **kwargs,
    ) -> None:
        """
        Initialize a .csv dataset.

        Args:
            filepath (str): path to .csv file.
            feature_list (FeatureList): a list of features. Defaults to None.
            pandas_dtype (str, type, dict): Optional parameter added to
                kwargs (and passed to pd.read_csv) as 'dtype'. Defaults to
                    None.
            kwargs (dict): additional parameters for pd.read_csv.

        """
        self.filepath = filepath
        self.initial_feature_list = feature_list  # see `setup_datasource`
        self.min_max_scaler = MinMaxScaler()
        self.standardizer = StandardScaler()
        self.kwargs = copy.deepcopy(kwargs)
        self.kwargs['dtype'] = pandas_dtype

        self.preprocess_df = (
            # may be applied to many chunks, so logic is determined once here
            self._reindex
            if self.initial_feature_list
            else self._id
        )
        self.setup_datasource()

        # self.notna_count
        self.max = self.min_max_scaler.data_max_
        self.min = self.min_max_scaler.data_min_
        self.mean = self.standardizer.mean_
        self.std = self.standardizer.scale_

    def _reindex(self, df: pd.DataFrame) -> pd.DataFrame:
        """Ensure given order of features, creating NaN columns for missing."""
        return df.reindex(columns=self.initial_feature_list, fill_value=np.NaN)

    def _id(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def setup_datasource(self):
        """
        Setup the datasource and compute statistics.

        The dataframe is read, calling `self.preprocess_df` on it and setting
        up the data as source, collecting statistics of the data.

        NOTE:
        To read item with a different subset and order of features,
        use the function returned by `get_feature_fn` or use the
        feature_mapping to ensure integer indexing.

        Sets:
        feature_list (FeatureList): feature names in this datasource.
        feature_mapping (pd.Series): maps feature name to index in items.
        feature_fn (CallableOnSource): function that indexes datasource with
            the feature_list.
        """
        raise NotImplementedError

    def get_feature_fn(self, feature_list: FeatureList) -> CallableOnSource:
        """Provides datasource specific indexing.

        Args:
            feature_list (FeatureList): subset of features to return in order.

        Returns:
            CallableOnSource: function that indexes datasource with the
                feature_list.
        """
        raise NotImplementedError

    def transform_datasource(
        self,
        transform_fn: CallableOnSource,
        feature_fn: CallableOnSource,
        impute: Optional[float] = None,
    ) -> None:
        """Apply scaling to the datasource.

        Args:
            transform_fn (CallableOnSource): transformation on source data.
            feature_fn (CallableOnSource): function that indexes datasource.
            impute (Optional[float]): NaN imputation with value if
                given. Defaults to None.
        """
        raise NotImplementedError

    def transform_dataset(
        self,
        transform_fn: CallableOnSource,
        feature_list: FeatureList,
        impute: Optional[float] = None,
    ) -> None:
        """Apply filtering, ordering and scaling to the datasource and update
        the dataset accordingly.

        Args:
            transform_fn (CallableOnSource): data scaling function, possibly
                feature wise.
            feature_list (FeatureList): subset of features to return in order.
            impute (Optional[float]): NaN imputation with value if
                given. Defaults to None.

        Sets:
        feature_list (FeatureList): feature names in this datasource.
        feature_mapping (pd.Series): maps feature name to index in items.
        feature_fn (CallableOnSource): function that indexes datasource with
            the feature_list.
        """
        # feature_fn for current state of datasource
        feature_fn = self.get_feature_fn(feature_list)
        statistics_indices = self.feature_mapping[feature_list].values  # noqa
        # update datasource
        self.transform_datasource(transform_fn, feature_fn, impute)
        # update statistics ordering (this could allow inverse transformation)
        self.max = self.max[statistics_indices]
        self.min = self.min[statistics_indices]
        self.mean = self.mean[statistics_indices]
        self.std = self.std[statistics_indices]
        self.notna_count = self.notna_count[statistics_indices]
        # delete outdated scalers
        try:
            del self.min_max_scaler
            del self.standardizer
        except NameError:
            pass
        # update dataset reflecting the transformation
        self.feature_list = feature_list
        self.feature_mapping = pd.Series(
            OrderedDict(
                [(feature, index) for index, feature in enumerate(self.feature_list)]
            )
        )
        self.feature_fn = self.get_feature_fn(self.feature_list)

    def __len__(self) -> int:
        raise NotImplementedError


def reduce_csv_statistics(
    csv_datasets: List[_CsvStatistics],
    feature_list: FeatureList = None,
) -> Tuple[FeatureList, np.array, np.array, np.array, np.array]:
    """
    Reduce datasets statistics.

    Args:
        csv_datasets (List[_CsvStatistics]): list of .csv datasets.
        feature_list (FeatureList): a list of features important to guarantee
            feature order preservation when multiple datasets are passed.
            Defaults to None, where features are string sorted.
    Returns:
        Tuple[FeatureList, np.array, np.array, np.array, np.array]: updated
            statistics with the following components:
                features (FeatureList): List of common, ordered features.
                maximum (np.array): Maximum per feature.
                minimum (np.array): Minimum per feature.
                mean (np.array): Mean per feature.
                std (np.array): Standard deviation per feature.

    """
    # collected common features
    features = list(
        reduce(
            lambda a_set, another_set: a_set & another_set,
            [set(csv_dataset.feature_mapping.keys()) for csv_dataset in csv_datasets],
        )
    )

    if feature_list is not None:
        # to sort features by key
        feature_ordering = {
            feature: index for index, feature in enumerate(feature_list)
        }
        features = sorted(features, key=lambda feature: feature_ordering[feature])
    else:
        # sorting the strings
        features = sorted(features)

    # features is the definite features_list

    maximums, minimums, means, stds, sample_numbers = zip(
        *[
            (
                # NOTE: here we ensure that we pick the statistics
                # in the right order via indexes provided by the dataset
                csv_dataset.max[csv_dataset.feature_mapping[features].values],
                csv_dataset.min[csv_dataset.feature_mapping[features].values],
                csv_dataset.mean[csv_dataset.feature_mapping[features].values],
                csv_dataset.std[csv_dataset.feature_mapping[features].values],
                # len(csv_dataset)
                csv_dataset.notna_count[csv_dataset.feature_mapping[features].values],
            )
            for csv_dataset in csv_datasets
        ]
    )
    # NOTE: reduce max and min
    maximum = np.nanmax(maximums, axis=0)
    minimum = np.nanmin(minimums, axis=0)
    # NOTE: reduce the mean
    total_number_of_samples = sum(sample_numbers).astype(float)
    mean = (
        np.nansum(
            [
                dataset_mean * number_of_samples
                for dataset_mean, number_of_samples in zip(means, sample_numbers)
            ],
            axis=0,
        )
        / total_number_of_samples
    )
    # NOTE: reduce the std
    std = np.sqrt(
        np.nansum(
            [
                (dataset_std**2 + dataset_mean**2) * number_of_samples
                for dataset_std, dataset_mean, number_of_samples in zip(
                    stds, means, sample_numbers
                )
            ],
            axis=0,
        )
        / total_number_of_samples
        - mean**2
    )
    return (features, maximum, minimum, mean, std)



"""Implementation of _CsvLazyDataset."""
from collections import OrderedDict

import numpy as np
import pandas as pd

from ..types import CallableOnSource, FeatureList, Hashable, Optional
from ._cache_datasource import _CacheDatasource
from ._csv_statistics import _CsvStatistics
from .base_dataset import KeyDataset


class _CsvLazyDataset(KeyDataset, _CacheDatasource, _CsvStatistics):
    """
    .csv dataset using lazy loading.

    Suggested when handling datasets that can't fit in the device memory.
    In case of datasets fitting in device memory consider using
    _CsvEagerDataset for better performance.
    """

    def __init__(
        self,
        filepath: str,
        feature_list: FeatureList = None,
        chunk_size: int = 10000,
        **kwargs,
    ) -> None:
        """
        Initialize a .csv dataset.

        Args:
            filepath (str): path to .csv file.
            feature_list (FeatureList): a list of features. Defaults to None.
            chunk_size (int): size of the chunks. Defaults to 10000.
            kwargs (dict): additional parameters for pd.read_csv.
                The argument chunksize is ignored in favor of chunck_size.
        """
        self.chunk_size = chunk_size

        _CacheDatasource.__init__(self, fit_size_limit_filepath=filepath)
        _CsvStatistics.__init__(
            self, filepath, feature_list=feature_list, **kwargs
        )  # calls setup_datasource
        _ = self.kwargs.pop('chunksize', None)  # not passing chunksize twice
        KeyDataset.__init__(self)

    def setup_datasource(self):
        """
        Setup the datasource and compute statistics.

        The dataframe is read, calling `self.preprocess_df` on it and setting
        up the data as source, collecting statistics of the data.

        NOTE:
        To read item with a different subset and order of features,
        use the function returned by `get_feature_fn` or use the
        feature_mapping to ensure integer indexing.

        Sets:
        feature_list (FeatureList): feature names in this datasource.
        feature_fn (CallableOnSource): function that indexes datasource with
            the feature_list.
        feature_mapping (pd.Series): maps feature name to index in items.
        """
        self.key_to_index_mapping = {}
        index = 0
        self.ordered_keys = []
        # look ahead for row length
        chunk = next(pd.read_csv(self.filepath, chunksize=1, **self.kwargs))
        self.notna_count = np.array([0] * self.preprocess_df(chunk).shape[1])

        for chunk in pd.read_csv(
            self.filepath, chunksize=self.chunk_size, **self.kwargs
        ):
            chunk = self.preprocess_df(chunk)
            self.min_max_scaler.partial_fit(chunk.values)
            self.standardizer.partial_fit(chunk.values)
            self.notna_count += chunk.notna().sum().values
            for key, row in chunk.iterrows():
                self.cache[index] = row.values
                self.key_to_index_mapping[key] = index
                index += 1
                self.ordered_keys.append(key)

        self.number_of_samples = len(self.ordered_keys)

        self.feature_list = chunk.columns.tolist()
        self.feature_mapping = pd.Series(
            OrderedDict(
                [(feature, index) for index, feature in enumerate(self.feature_list)]
            )
        )
        self.feature_fn = self.get_feature_fn(self.feature_list)

    def get_feature_fn(self, feature_list: FeatureList) -> CallableOnSource:
        """Provides datasource specific indexing.

        Args:
            feature_list (FeatureList): subset of features to return in order.

        Returns:
            CallableOnSource: function that indexes datasource with the
                feature_list.
        """
        indices = self.feature_mapping[feature_list].values

        def feature_fn(sample: np.ndarray) -> np.ndarray:
            return sample[indices]

        return feature_fn

    def transform_datasource(
        self,
        transform_fn: CallableOnSource,
        feature_fn: CallableOnSource,
        impute: Optional[float] = None,
    ) -> None:
        """Apply scaling to the datasource.

        Args:
            transform_fn (CallableOnSource): transformation on source data.
            feature_fn (CallableOnSource): function that indexes datasource.
            impute (Optional[float]): NaN imputation with value if
                given. Defaults to None.
        """
        if impute is None:
            for index in self.cache:
                self.cache[index] = transform_fn(feature_fn(self.cache[index]))
        else:
            for index in self.cache:
                transformed = transform_fn(feature_fn(self.cache[index]))
                self.cache[index] = np.nan_to_num(transformed, impute)

    def __len__(self) -> int:
        """Total number of samples."""
        return self.number_of_samples

    def __getitem__(self, index: int) -> np.array:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            np.array: the current sample read from cache.
        """
        return self.cache[index]

    def get_key(self, index: int) -> Hashable:
        """Get key from integer index."""
        return self.ordered_keys[index]

    def get_index(self, key: Hashable) -> int:
        """Get index for first datum mapping to the given key."""
        return self.key_to_index_mapping[key]

    def keys(self):
        return iter(self.ordered_keys)

    @property
    def has_duplicate_keys(self):
        return self.number_of_samples != len(self.key_to_index_mapping)

    def __del__(self):
        """Delete the _CsvLazyDataset."""
        _CacheDatasource.__del__(self)



"""Testing Protein Sequence dataset."""
import os
import random
import time
import unittest

import numpy as np
from importlib_resources import files
from torch.utils.data import DataLoader

from pytoda.datasets import ProteinSequenceDataset
from pytoda.tests.utils import TestFileContent

SMI_CONTENT = os.linesep.join(['EGK	ID3', 'S	ID1', 'FGAAV	ID2', 'NCCS	ID4'])
MORE_SMI_CONTENT = os.linesep.join(['KGE	ID5', 'K	ID6', 'SCCN	ID7', 'K	ID8'])
BROKEN_SMI_CONTENT = os.linesep.join(['KGE	ID5', 'K	ID6', 'SCfCN	ID7'])
AS_SMI_CONTENT = os.linesep.join(
    [
        'LGQGTRTNVVKTMLAVMVTEYVEHGPVLVRNLSDV	P29597',
        'LGKGTFGKVAKELLTLFVMEYANGGEFVVENMTDL	P31751',
    ]
)
AS_MORE_SMI_CONTENT = os.linesep.join(
    [
        'IGEGSTGIVAKEMVVMVVMEFLEGGADVIDSLSDF	Q9P286',
        'LGKGTFGKVAKELLTLFVMEYANGGEFVVENMTDL	P31749',  # same AS different protein
    ]
)


FASTA_CONTENT_UNIPROT = r""">sp|Q6GZX0|005R_FRG3G Uncharacterized protein 005R OS=Frog virus 3 (isolate Goorha) OX=654924 GN=FV3-005R PE=4 SV=1
MQNPLPEVMSPEHDKRTTTPMSKEANKFIRELDKKPGDLAVVSDFVKRNTGKRLPIGKRS
NLYVRICDLSGTIYMGETFILESWEELYLPEPTKMEVLGTLESCCGIPPFPEWIVMVGED
QCVYAYGDEEILLFAYSVKQLVEEGIQETGISYKYPDDISDVDEEVLQQDEEIQKIRKKT
REFVDKDAQEFQDFLNSLDASLLS
>sp|Q91G88|006L_IIV6 Putative KilA-N domain-containing protein 006L OS=Invertebrate iridescent virus 6 OX=176652 GN=IIV6-006L PE=3 SV=1
MDSLNEVCYEQIKGTFYKGLFGDFPLIVDKKTGCFNATKLCVLGGKRFVDWNKTLRSKKL
IQYYETRCDIKTESLLYEIKGDNNDEITKQITGTYLPKEFILDIASWISVEFYDKCNNII
"""  # length 204, 120

FASTA_CONTENT_GENERIC = (
    FASTA_CONTENT_UNIPROT
    + r""">generic_header eager upfp would concat to sequence above.
LLLLLLLLLLLLLLLL
"""
)  # length 16

all_keys = ['ID3', 'ID1', 'ID2', 'ID4', 'Q6GZX0', 'Q91G88']


class TestProteinSequenceDatasetEagerBackend(unittest.TestCase):
    """Testing ProteinSequence dataset with eager backend."""

    def setUp(self):
        self.backend = 'eager'
        print(f'backend is {self.backend}')
        self.smi_content = SMI_CONTENT
        self.smi_other_content = MORE_SMI_CONTENT
        self.smi_broken_content = BROKEN_SMI_CONTENT
        self.as_smi_content = AS_SMI_CONTENT
        self.as_more_smi_content = AS_MORE_SMI_CONTENT
        # would fail with FASTA_CONTENT_GENERIC
        self.fasta_content = FASTA_CONTENT_UNIPROT

    def test___len__smi(self) -> None:
        """Test __len__."""

        with TestFileContent(self.smi_content) as a_test_file:
            with TestFileContent(self.smi_other_content) as another_test_file:
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                )
                self.assertEqual(len(protein_sequence_dataset), 8)

    def test___len__fasta(self) -> None:
        """Test __len__."""
        with TestFileContent(self.fasta_content) as a_test_file:
            protein_sequence_dataset = ProteinSequenceDataset(
                a_test_file.filename, filetype='.fasta', backend=self.backend
            )
            # eager only uniprot headers
            self.assertEqual(len(protein_sequence_dataset), 2)
            time.sleep(1)

    def test___getitem__(self) -> None:
        """Test __getitem__."""
        with TestFileContent(self.smi_content) as a_test_file:
            with TestFileContent(self.smi_other_content) as another_test_file:
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=True,
                    add_start_and_stop=True,
                    backend=self.backend,
                )
                pad_index = protein_sequence_dataset.protein_language.token_to_index[
                    '<PAD>'
                ]
                start_index = protein_sequence_dataset.protein_language.token_to_index[
                    '<START>'
                ]
                stop_index = protein_sequence_dataset.protein_language.token_to_index[
                    '<STOP>'
                ]
                e_index = protein_sequence_dataset.protein_language.token_to_index['E']
                g_index = protein_sequence_dataset.protein_language.token_to_index['G']
                k_index = protein_sequence_dataset.protein_language.token_to_index['K']
                n_index = protein_sequence_dataset.protein_language.token_to_index['N']
                c_index = protein_sequence_dataset.protein_language.token_to_index['C']
                s_index = protein_sequence_dataset.protein_language.token_to_index['S']

                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(),
                    [
                        pad_index,
                        pad_index,
                        start_index,
                        e_index,
                        g_index,
                        k_index,
                        stop_index,
                    ],
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(),
                    [
                        pad_index,
                        start_index,
                        n_index,
                        c_index,
                        c_index,
                        s_index,
                        stop_index,
                    ],
                )
                self.assertListEqual(
                    protein_sequence_dataset[7].numpy().flatten().tolist(),
                    [
                        pad_index,
                        pad_index,
                        pad_index,
                        pad_index,
                        start_index,
                        k_index,
                        stop_index,
                    ],
                )

                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=False,
                    backend=self.backend,
                )
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(),
                    [e_index, g_index, k_index],
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(),
                    [n_index, c_index, c_index, s_index],
                )
                self.assertListEqual(
                    protein_sequence_dataset[7].numpy().flatten().tolist(), [k_index]
                )

                # Test padding but no start and stop token
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=True,
                    add_start_and_stop=False,
                    backend=self.backend,
                )
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(),
                    [pad_index, pad_index, e_index, g_index, k_index],
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(),
                    [pad_index, n_index, c_index, c_index, s_index],
                )
                self.assertListEqual(
                    protein_sequence_dataset[7].numpy().flatten().tolist(),
                    [pad_index, pad_index, pad_index, pad_index, k_index],
                )

                # Test augmentation / order reversion
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    augment_by_revert=True,
                )

                random.seed(42)
                for reverted_sequence in ['EGK', 'KGE', 'KGE', 'KGE']:
                    token_indexes = (
                        protein_sequence_dataset[0].numpy().flatten().tolist()
                    )
                    sequence = protein_sequence_dataset.protein_language.token_indexes_to_sequence(
                        token_indexes
                    )
                    self.assertEqual(sequence, reverted_sequence)
                for reverted_sequence in ['S', 'S', 'S', 'S']:
                    token_indexes = (
                        protein_sequence_dataset[1].numpy().flatten().tolist()
                    )
                    sequence = protein_sequence_dataset.protein_language.token_indexes_to_sequence(
                        token_indexes
                    )
                    self.assertEqual(sequence, reverted_sequence)

                # Test UNIREP vocab
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    amino_acid_dict='unirep',
                    padding=True,
                    add_start_and_stop=False,
                    backend=self.backend,
                )
                pad_index = protein_sequence_dataset.protein_language.token_to_index[
                    '<PAD>'
                ]
                e_index = protein_sequence_dataset.protein_language.token_to_index['E']
                g_index = protein_sequence_dataset.protein_language.token_to_index['G']
                k_index = protein_sequence_dataset.protein_language.token_to_index['K']
                n_index = protein_sequence_dataset.protein_language.token_to_index['N']
                c_index = protein_sequence_dataset.protein_language.token_to_index['C']
                s_index = protein_sequence_dataset.protein_language.token_to_index['S']
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(),
                    [pad_index, pad_index, e_index, g_index, k_index],
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(),
                    [pad_index, n_index, c_index, c_index, s_index],
                )
                self.assertListEqual(
                    protein_sequence_dataset[7].numpy().flatten().tolist(),
                    [pad_index, pad_index, pad_index, pad_index, k_index],
                )

        # Test parsing of .fasta file
        with TestFileContent(self.fasta_content) as a_test_file:
            protein_sequence_dataset = ProteinSequenceDataset(
                a_test_file.filename,
                filetype='.fasta',
                add_start_and_stop=True,
                backend=self.backend,
            )
            a_tokenized_sequence = protein_sequence_dataset[1].tolist()
            self.assertEqual(len(a_tokenized_sequence), 206)
            # padded to length + start + stop
            self.assertEqual(sum(a_tokenized_sequence[:-123]), 0)
            time.sleep(1)

        # Test case with unknown token in dataset
        for iterate in [False, True]:
            with TestFileContent(self.smi_broken_content) as a_test_file:
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    add_start_and_stop=False,
                    padding=False,
                    backend=self.backend,
                    iterate_dataset=iterate,
                )
                self.assertListEqual(
                    protein_sequence_dataset[2].tolist(),
                    [
                        protein_sequence_dataset.protein_language.token_to_index['S'],
                        protein_sequence_dataset.protein_language.token_to_index['C'],
                        protein_sequence_dataset.protein_language.token_to_index[
                            '<UNK>'
                        ],
                        protein_sequence_dataset.protein_language.token_to_index['C'],
                        protein_sequence_dataset.protein_language.token_to_index['N'],
                    ],
                )

        """
        With sequence augmentation strategies
        """
        # TODO: Test swapping, test noise, test custom file
        with TestFileContent(self.as_smi_content) as a_test_file:
            with TestFileContent(self.as_more_smi_content) as another_test_file:

                # No *actual* changes
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=False,
                    backend=self.backend,
                    sequence_augment={'discard_lowercase': True},
                )
                p_to_t = (
                    protein_sequence_dataset.protein_language.sequence_to_token_indexes
                )
                gts = [
                    p_to_t('LGQGTRTNVVKTMLAVMVTEYVEHGPVLVRNLSDV'),
                    p_to_t('LGKGTFGKVAKELLTLFVMEYANGGEFVVENMTDL'),
                ]

                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(), gts[0]
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(), gts[1]
                )

                # Converted to full sequence
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=False,
                    backend=self.backend,
                    sequence_augment={'discard_lowercase': False},
                )
                gts = [
                    p_to_t(
                        'ITQLSHLGQGTRTNVYEGRLRVEGSGDPEEGKMDDEDPLVPGRDRGQELRVVLKVLDPSHHDIALAFYETASLMSQVSHTHLAFVHGVCVRGPENIMVTEYVEHGPLDVWLRRERGHVPMAWKMVVAQQLASALSYLENKNLVHGNVCGRNILLARLGLAEGTSPFIKLSDPGVGLGALSREERVERIPWLAPECLPGGANSLSTAMDKWGFGATLLEICFDGEAPLQSRSPSEKEHFYQRQHRLPEPSCPQLATLTSQCLTYEPTQRPSFRTILRDLTR'
                    ),
                    p_to_t(
                        'FEYLKLLGKGTFGKVILVKEKATGRYYAMKILKKEVIVAKDEVAHTLTENRVLQNSRHPFLTALKYSFQTHDRLCFVMEYANGGELFFHLSRERVFSEDRARFYGAEIVSALDYLHSEKNVVYRDLKLENLMLDKDGHIKITDFGLCKEGIKDGATMKTFCGTPEYLAPEVLEDNDYGRAVDWWGLGVVMYEMMCGRLPFYNQDHEKLFELILMEEIRFPRTLGPEAKSLLSGLLKKDPKQRLGGGSEDAKEIMQHRFF'
                    ),
                ]
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(), gts[0]
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(), gts[1]
                )

                # Flip substrings
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=False,
                    backend=self.backend,
                    sequence_augment={
                        'discard_lowercase': True,
                        'flip_substrings': 1.0,
                    },
                )
                gts = [
                    p_to_t('VNTRTGQGLVKTMALVPGHEVYETVMVVLNRLDSV'),
                    p_to_t('VKGFTGKGLAKELTLLEGGNAYEMVFFVVNEMDTL'),
                ]
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(), gts[0]
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(), gts[1]
                )

                # Swap substrings
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=False,
                    backend=self.backend,
                    sequence_augment={
                        'discard_lowercase': True,
                        'swap_substrings': 1.0,
                    },
                )
                gts = [
                    p_to_t('VKTMLAVMVTEYVEHGPVLVRNLSDVLGQGTRTNV'),
                    p_to_t('AKELLTLFVMEYANGGEFVVENMTDLLGKGTFGKV'),
                ]
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(), gts[0]
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(), gts[1]
                )

                # Add noise (on active site)
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=False,
                    backend=self.backend,
                    sequence_augment={
                        'discard_lowercase': True,
                        'swap_substrings': 1.0,
                        'noise': (0.5, 0.0),
                    },
                )
                gts = [
                    p_to_t('VKMMLQEPVPWCVGNMHKNCKNLSDVDGQGTRAQY'),
                    p_to_t('AGEHWTSFVMDFANGQKPCVCNMTDDDGHAHFVKV'),
                ]
                np.random.seed(42)
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(), gts[0]
                )
                np.random.seed(42)
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(), gts[1]
                )

                # Add noise (outside active site)
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=False,
                    backend=self.backend,
                    sequence_augment={
                        'discard_lowercase': False,
                        'noise': (0.0, 0.1),
                    },
                )
                gts = [
                    p_to_t(
                        'ITWLSQLGQGTRTNVYEGRLRVEGSGDIEEGKMDDEDKLVMGRDRGQELRFVLKVLDPSHHDIPLAFYETASLMSQVSHTHLAFVHGVCVPGIENIMVTEYVEHGPLDVWLRAERGHVPTAWKMVAAQQLASALSKLENKNLVHGNVCGRNILLARLGLAEITSPFIKLSDPGVQLGALSRWERVERIPWLAPECLPNGANSTSTAADKWGFGRTLLEICFDGEAPLQSRSPSEKEWFYQRYRRLPEPSPPQLATLTWQFLKYAPTIRPSFRTSLRDRQR'
                    ),
                    p_to_t(
                        'FEWLKQLGKGTFGKVILVKEKATGRYYAMKILKKEVIVAHDEVWHTLHENRVLQNSRHPFLTALKRSPQTHDRLCFVMEYANGGELFFHLSRERPFIEDRARFYGAYIVSALDYLASEKNVVTRDLKLENLMYDKDGHKKITDFGLCKEGIKDNATMKTFCGTPEYLAPEVLEDNDYQRAVDWWWLGVVMYEMMCGRLPFNNQDHTKLFALILMEERRFPRTLGPEAKSLLSGLLKKDPWQRLGYRSEDAKEPMQHRFF'
                    ),
                ]
                np.random.seed(42)
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(), gts[0]
                )

                np.random.seed(42)
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(), gts[1]
                )

                # Test passing a file
                path = files('pytoda.proteins.metadata').joinpath(
                    'kinase_activesite_alignment.smi'
                )
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    padding=False,
                    add_start_and_stop=False,
                    backend=self.backend,
                    sequence_augment={
                        'alignment_path': path,
                        'discard_lowercase': True,
                        'flip_substrings': 1.0,
                    },
                )
                gts = [
                    p_to_t('VNTRTGQGLVKTMALVPGHEVYETVMVVLNRLDSV'),
                    p_to_t('VKGFTGKGLAKELTLLEGGNAYEMVFFVVNEMDTL'),
                ]
                self.assertListEqual(
                    protein_sequence_dataset[0].numpy().flatten().tolist(), gts[0]
                )
                self.assertListEqual(
                    protein_sequence_dataset[3].numpy().flatten().tolist(), gts[1]
                )

    def test_data_loader(self) -> None:
        """Test data_loader."""
        with TestFileContent(self.smi_content) as a_test_file:
            with TestFileContent(self.smi_other_content) as another_test_file:
                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    add_start_and_stop=False,
                    backend=self.backend,
                )
                data_loader = DataLoader(
                    protein_sequence_dataset, batch_size=4, shuffle=True
                )
                for batch_index, batch in enumerate(data_loader):
                    self.assertEqual(batch.shape, (4, 5))
                    if batch_index > 10:
                        break

                protein_sequence_dataset = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    add_start_and_stop=True,
                    backend=self.backend,
                )
                data_loader = DataLoader(
                    protein_sequence_dataset, batch_size=4, shuffle=True
                )
                for batch_index, batch in enumerate(data_loader):
                    self.assertEqual(batch.shape, (4, 7))
                    if batch_index > 10:
                        break

    def _test_indexed(self, ds, keys, index):
        key = keys[index]
        positive_index = index % len(ds)
        # get_key (support for negative index?)
        self.assertEqual(key, ds.get_key(positive_index))
        self.assertEqual(key, ds.get_key(index))
        # get_index
        self.assertEqual(positive_index, ds.get_index(key))
        # get_item_from_key
        self.assertTrue(all(ds[index] == ds.get_item_from_key(key)))
        # keys
        self.assertSequenceEqual(keys, list(ds.keys()))
        # duplicate keys
        self.assertFalse(ds.has_duplicate_keys)

    def test_all_base_for_indexed_methods(self):

        with TestFileContent(self.smi_content) as a_test_file:
            with TestFileContent(self.smi_other_content) as another_test_file:
                protein_sequence_ds = ProteinSequenceDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                )
                protein_sequence_ds_0 = ProteinSequenceDataset(
                    a_test_file.filename, backend=self.backend
                )
                protein_sequence_ds_1 = ProteinSequenceDataset(
                    another_test_file.filename, backend=self.backend
                )
        all_smiles, all_keys = zip(
            *(
                pair.split('\t')
                for pair in (
                    self.smi_content.split(os.linesep)
                    + self.smi_other_content.split(os.linesep)
                )
            )
        )

        for ds, keys in [
            (protein_sequence_ds, all_keys),
            (protein_sequence_ds_0, all_keys[:4]),
            (protein_sequence_ds_1, all_keys[4:]),
            # no transformation on
            # concat delegation to _SmiLazyDataset/_SmiEagerDataset
            (protein_sequence_ds_0 + protein_sequence_ds_1, all_keys),
        ]:
            index = -1
            self._test_indexed(ds, keys, index)

        # duplicate
        duplicate_ds = protein_sequence_ds_0 + protein_sequence_ds_0
        self.assertTrue(duplicate_ds.has_duplicate_keys)

        # ProteinSequenceDataset tests and raises
        with TestFileContent(self.smi_content) as a_test_file:
            with self.assertRaises(KeyError):
                protein_sequence_ds = ProteinSequenceDataset(
                    a_test_file.filename, a_test_file.filename, backend=self.backend
                )


class TestProteinSequenceDatasetLazyBackend(
    TestProteinSequenceDatasetEagerBackend
):  # noqa
    """Testing ProteinSequence dataset with lazy backend."""

    def setUp(self):
        self.backend = 'lazy'
        print(f'backend is {self.backend}')
        self.smi_content = SMI_CONTENT
        self.smi_other_content = MORE_SMI_CONTENT
        self.smi_broken_content = BROKEN_SMI_CONTENT
        self.fasta_content = FASTA_CONTENT_GENERIC
        self.as_smi_content = AS_SMI_CONTENT
        self.as_more_smi_content = AS_MORE_SMI_CONTENT

    def test___len__fasta(self) -> None:
        """Test __len__."""
        with TestFileContent(self.fasta_content) as a_test_file:
            protein_sequence_dataset = ProteinSequenceDataset(
                a_test_file.filename, filetype='.fasta', backend=self.backend
            )
            # generic sequences
            self.assertEqual(len(protein_sequence_dataset), 3)
            time.sleep(1)


if __name__ == '__main__':
    unittest.main()



"""Testing transforms."""
import random
import unittest

import torch

from pytoda.transforms import (
    AugmentByReversing,
    Compose,
    LeftPadding,
    ListToTensor,
    ToTensor,
)


class TestTransforms(unittest.TestCase):
    """Testing transforms."""

    def test_left_padding(self) -> None:
        """Test LeftPadding."""

        padding_index = 0
        padding_lengths = [8, 4]

        # Molecules that are too long will be cut and a warning will be raised.
        for padding_length in padding_lengths:
            transform = LeftPadding(
                padding_index=padding_index, padding_length=padding_length
            )
            for mol in ['C(N)CS', 'CCO']:
                self.assertEqual(len(transform(list(mol))), padding_length)

    def test_augment_by_reversing(self) -> None:
        """Test AugmentByReversing."""

        sequence = 'ABC'
        ground_truths = ['ABC', 'ABC', 'CBA']
        for k in range(15):
            for p, ground_truth in zip([0.0, 0.5, 1.0], ground_truths):
                random.seed(42)
                transform = AugmentByReversing(p=p)
                self.assertEqual(transform(sequence), ground_truth)

    def test_to_tensor(self) -> None:
        """Test ToTensor."""

        tokens = [2, 3, 4]
        transform = ToTensor()
        tensor = transform(tokens)
        self.assertListEqual(
            [tokens[0], tokens[1], tokens[2]], [tensor[0], tensor[1], tensor[2]]
        )
        self.assertTrue(torch.is_tensor(tensor))
        self.assertEqual(len(tensor), 3)
        self.assertEqual(len(tensor.shape), 1)
        self.assertRaises(TypeError, ToTensor, dtype=42)

    def test_list_to_tensor(self) -> None:
        """Test ListToTensor."""

        tokens = [(2, 3, 4), (2, 3, 4)]
        transform = ListToTensor()
        tensor = transform(tokens)
        self.assertEqual(len(tensor), 2)
        self.assertEqual(len(tensor.shape), 2)
        self.assertEqual(tensor.shape[-1], 3)
        self.assertListEqual(
            [tokens[0][0], tokens[0][1], tokens[0][2]],
            [tensor[0][0], tensor[0][1], tensor[0][2]],
        )
        self.assertTrue(torch.is_tensor(tensor))
        self.assertRaises(TypeError, ListToTensor, dtype=42)

    def test_compose(self) -> None:
        """Test Compose."""

        # Test equality
        c1 = Compose([ToTensor()])
        c2 = Compose([ToTensor()])
        c3 = Compose([LeftPadding(padding_length=2, padding_index=0), ToTensor()])
        c4 = Compose([ToTensor(dtype=torch.long)])
        self.assertTrue(c1 == c2)
        self.assertFalse(c1 == c3)
        self.assertFalse(c1 == c4)

        # Test repr
        self.assertIsNotNone(repr(c1))


if __name__ == '__main__':
    unittest.main()
